part of barcodes;

/// The [Codabar] is a discrete numeric symbology that can encode 0-9 digits, six symbols, and
/// plus an additional 4 start and stop characters.
class Codabar extends Symbology {
  /// Create a [Codabar] symbology with the default or required properties.
  ///
  /// The arguments [module] must be non-negative and greater than 0.
  ///
  Codabar({int module}) : super(module: module) {
    _codeBarMap = <String, String>{
      '0': '101010011',
      '1': '101011001',
      '2': '101001011',
      '3': '110010101',
      '4': '101101001',
      '5': '110101001',
      '6': '100101011',
      '7': '100101101',
      '8': '100110101',
      '9': '110100101',
      '-': '101001101',
      '\$': '101100101',
      ':': '1101011011',
      '/': '1101101011',
      '.': '1101101101',
      '+': '101100110011',
      'A': '1011001001',
      'B': '1001001011',
      'C': '1010010011',
      'D': '1010011001'
    };
  }

  /// Represents the supported symbol and its byte value
  Map<String, String> _codeBarMap;

  @override
  bool _validateInput(String value) {
    for (int i = 0; i < value.length; i++) {
      if (!_codeBarMap.containsKey(value[i]) ||
          value[i] == 'A' ||
          value[i] == 'B' ||
          value[i] == 'C' ||
          value[i] == 'D') {
        throw 'The provided input cannot be encoded : ' + value[i];
      }
    }
    return true;
  }

  @override
  void _renderBarcode(
      Canvas canvas,
      Size size,
      Offset offset,
      String value,
      Color foregroundColor,
      TextStyle textStyle,
      double textSpacing,
      TextAlign textAlign,
      bool showValue) {
    final Paint _paint = _getBarPaint(foregroundColor);
    final List<String> code = _getCodeValues(value);
    final int _barTotalLength = _calculateTotalLength(code);
    double _left = module == null
        ? offset.dx
        : _calculateLeftPosition(
            _barTotalLength, module, size.width, offset.dx);
    final Rect _barCodeRect = Rect.fromLTRB(
        offset.dx, offset.dy, offset.dx + size.width, offset.dy + size.height);
    double _ratio = 0;
    if (module != null) {
      _ratio = module.toDouble();
    } else {
      // Calculates the bar length based on number of individual bar codes
      final int singleModule = (size.width ~/ _barTotalLength).toInt();
      _ratio = singleModule.toDouble();
      final double _leftPadding = (size.width - (_barTotalLength * _ratio)) / 2;
      _left += _leftPadding;
    }
    _left = _left.roundToDouble();
    for (int i = 0; i < code.length; i++) {
      final String _codeValue = code[i];
      const bool _hasExtraHeight = false;
      final double _barHeight = _hasExtraHeight
          ? size.height + _textSize.height + textSpacing
          : size.height;
      final int _codeLength = _codeValue.length;
      for (int j = 0; j < _codeLength; j++) {
        final bool _canDraw = _codeValue[j] == '1' ? true : false;

        // Draws the barcode when the corresponding bar value is one
        if (_canDraw &&
            (_left >= _barCodeRect.left &&
                _left + _ratio < _barCodeRect.right)) {
          final Rect _individualBarRect = Rect.fromLTRB(
              _left, offset.dy, _left + _ratio, offset.dy + _barHeight);
          canvas.drawRect(_individualBarRect, _paint);
        }
        _left += _ratio;
      }
      if (i < code.length - 1) {
        _left += _ratio;
      }
    }
    if (showValue)
      _drawText(canvas, offset, size, value, textStyle, textSpacing, textAlign);
  }

  /// Calculate total bar length from give input value
  int _calculateTotalLength(List<String> code) {
    int _count = 0;
    for (int i = 0; i < code.length; i++) {
      final int _numberOfDigits = code[i].length;
      _count += _numberOfDigits;
    }
    _count += code.length - 1;
    return _count;
  }

  /// Method to append the start and the stop symbol
  String _getValueWithStartAndStopSymbol(String value) {
    return 'A' + value + 'A';
  }

  /// Returns the encoded value of the provided input value
  List<String> _getCodeValues(String value) {
    _valueWithStartAndStopSymbol = _getValueWithStartAndStopSymbol(value);
    final List<String> _codeBarValues =
        List<String>(_valueWithStartAndStopSymbol.length);
    for (int i = 0; i < _valueWithStartAndStopSymbol.length; i++) {
      for (int j = 0; j < _codeBarMap.length; j++) {
        if (_valueWithStartAndStopSymbol[i] ==
            _codeBarMap.entries.elementAt(j).key) {
          _codeBarValues[i] = _codeBarMap.entries.elementAt(j).value;
          break;
        }
      }
    }
    return _codeBarValues;
  }
}
