part of barcodes;

/// The [DataMatrix] is two-dimensional barcode. The information to be encoded with text or numeric values.
/// Each barcode can store values up to 2335.
///
/// [DataMatrix] barcode consists of a grid of dark and light dots or blocks forming a square or rectangular symbol.
///
/// [DataMatrix] barcode will be mostly used for courier parcel, food industry, etc.
///
class DataMatrix extends Symbology {
  /// Create a [DataMatrix] symbology with the default or required properties.
  ///
  /// The arguments [module] must be non-negative and greater than 0.
  ///
  DataMatrix(
      {int module,
      this.dataMatrixSize = DataMatrixSize.auto,
      this.encoding = DataMatrixEncoding.auto})
      : super(module: module) {
    _initialize();
  }

  /// Define the size that is used to encode the amount of data.
  ///
  /// Defaults to [DataMatrixSize.auto].
  ///
  /// Also refer [DataMatrixSize].
  ///
  /// ```dart
  /// Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfBarcodeGenerator(value:'123456',
  ///        symbology: DataMatrix( dataMatrixSize: DataMatrixSize.size52x52,
  ///            )));
  ///}
  /// ```dart
  final DataMatrixSize dataMatrixSize;

  /// Defines the encoding type for the [DataMatrix] code.
  ///
  /// Multiple encoding modes are used to store different kinds of messages in [DataMatrix] barcode.
  ///
  /// When the encoding type is [DataMatrixEncoding.ascii], the codeword will be calculated
  /// as below,
  ///
  /// Codeword = ASCII value + 1, where ASCII value ranges from 0 to 127.
  ///
  /// When the encoding type is [DataMatrixEncoding.base256], the first codeword will be calculated with the value 235 and
  /// the second code value is calculated as ASCII value - 127.
  /// The base256 value ranges from 128 to 255.
  ///
  /// When the encoding type is [DataMatrixEncoding.asciiNumeric], then the codeword will be calculated
  /// as below,
  ///
  /// Codeword = numerical value pair + 130, where the numerical value pair will be like 00, 01, 02,.....99.
  ///
  /// Defaults to [DataMatrixEncoding.auto].
  ///
  /// Also refer [DataMatrixEncoding].
  ///
  /// ```dart
  /// Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfBarcodeGenerator(value:'123456',
  ///        symbology: DataMatrix(
  ///        encoding: DataMatrixEncoding.base256,
  //            )));
  ///}
  /// ```dart
  final DataMatrixEncoding encoding;

  /// Defines the list of symbol attributes
  List<_DataMatrixSymbolAttribute> _symbolAttributes;

  /// Defines the log list
  List<int> _log;

  /// Defines the aLog list
  List<int> _aLog;

  /// Defines the block length
  int _blockLength;

  /// Defines the polynomial collection
  List<int> _polynomial;

  /// Defines the symbol attributes
  _DataMatrixSymbolAttribute _symbolAttribute;

  /// Defines the actual encoding value
  DataMatrixEncoding _encoding;

  /// Defines the data matrix input value
  String _inputValue;

  /// Defines the list of data matrix
  List<List<int>> _dataMatrixList;

  /// Defines the list of encoded code words
  List<int> _encodedCodeword;

  /// Initializes the symbol attributes
  void _initialize() {
    _symbolAttributes = <_DataMatrixSymbolAttribute>[
      _DataMatrixSymbolAttribute(
          symbolRow: 10,
          symbolColumn: 10,
          horizontalDataRegion: 1,
          verticalDataRegion: 1,
          dataCodeWords: 3,
          correctionCodeWords: 5,
          interLeavedBlock: 1,
          interLeavedDataBlock: 3),
      _DataMatrixSymbolAttribute(
          symbolRow: 12,
          symbolColumn: 12,
          horizontalDataRegion: 1,
          verticalDataRegion: 1,
          dataCodeWords: 5,
          correctionCodeWords: 7,
          interLeavedBlock: 1,
          interLeavedDataBlock: 5),
      _DataMatrixSymbolAttribute(
          symbolRow: 14,
          symbolColumn: 14,
          horizontalDataRegion: 1,
          verticalDataRegion: 1,
          dataCodeWords: 8,
          correctionCodeWords: 10,
          interLeavedBlock: 1,
          interLeavedDataBlock: 8),
      _DataMatrixSymbolAttribute(
          symbolRow: 16,
          symbolColumn: 16,
          horizontalDataRegion: 1,
          verticalDataRegion: 1,
          dataCodeWords: 12,
          correctionCodeWords: 12,
          interLeavedBlock: 1,
          interLeavedDataBlock: 12),
      _DataMatrixSymbolAttribute(
          symbolRow: 18,
          symbolColumn: 18,
          horizontalDataRegion: 1,
          verticalDataRegion: 1,
          dataCodeWords: 18,
          correctionCodeWords: 14,
          interLeavedBlock: 1,
          interLeavedDataBlock: 18),
      _DataMatrixSymbolAttribute(
          symbolRow: 20,
          symbolColumn: 20,
          horizontalDataRegion: 1,
          verticalDataRegion: 1,
          dataCodeWords: 22,
          correctionCodeWords: 18,
          interLeavedBlock: 1,
          interLeavedDataBlock: 22),
      _DataMatrixSymbolAttribute(
          symbolRow: 22,
          symbolColumn: 22,
          horizontalDataRegion: 1,
          verticalDataRegion: 1,
          dataCodeWords: 30,
          correctionCodeWords: 20,
          interLeavedBlock: 1,
          interLeavedDataBlock: 30),
      _DataMatrixSymbolAttribute(
          symbolRow: 24,
          symbolColumn: 24,
          horizontalDataRegion: 1,
          verticalDataRegion: 1,
          dataCodeWords: 36,
          correctionCodeWords: 24,
          interLeavedBlock: 1,
          interLeavedDataBlock: 36),
      _DataMatrixSymbolAttribute(
          symbolRow: 26,
          symbolColumn: 26,
          horizontalDataRegion: 1,
          verticalDataRegion: 1,
          dataCodeWords: 44,
          correctionCodeWords: 28,
          interLeavedBlock: 1,
          interLeavedDataBlock: 44),
      _DataMatrixSymbolAttribute(
          symbolRow: 32,
          symbolColumn: 32,
          horizontalDataRegion: 2,
          verticalDataRegion: 2,
          dataCodeWords: 62,
          correctionCodeWords: 36,
          interLeavedBlock: 1,
          interLeavedDataBlock: 62),
      _DataMatrixSymbolAttribute(
          symbolRow: 36,
          symbolColumn: 36,
          horizontalDataRegion: 2,
          verticalDataRegion: 2,
          dataCodeWords: 86,
          correctionCodeWords: 42,
          interLeavedBlock: 1,
          interLeavedDataBlock: 86),
      _DataMatrixSymbolAttribute(
          symbolRow: 40,
          symbolColumn: 40,
          horizontalDataRegion: 2,
          verticalDataRegion: 2,
          dataCodeWords: 114,
          correctionCodeWords: 48,
          interLeavedBlock: 1,
          interLeavedDataBlock: 114),
      _DataMatrixSymbolAttribute(
          symbolRow: 44,
          symbolColumn: 44,
          horizontalDataRegion: 2,
          verticalDataRegion: 2,
          dataCodeWords: 144,
          correctionCodeWords: 56,
          interLeavedBlock: 1,
          interLeavedDataBlock: 144),
      _DataMatrixSymbolAttribute(
          symbolRow: 48,
          symbolColumn: 48,
          horizontalDataRegion: 2,
          verticalDataRegion: 2,
          dataCodeWords: 174,
          correctionCodeWords: 68,
          interLeavedBlock: 1,
          interLeavedDataBlock: 174),
      _DataMatrixSymbolAttribute(
          symbolRow: 52,
          symbolColumn: 52,
          horizontalDataRegion: 2,
          verticalDataRegion: 2,
          dataCodeWords: 204,
          correctionCodeWords: 84,
          interLeavedBlock: 2,
          interLeavedDataBlock: 102),
      _DataMatrixSymbolAttribute(
          symbolRow: 64,
          symbolColumn: 64,
          horizontalDataRegion: 4,
          verticalDataRegion: 4,
          dataCodeWords: 280,
          correctionCodeWords: 112,
          interLeavedBlock: 2,
          interLeavedDataBlock: 140),
      _DataMatrixSymbolAttribute(
          symbolRow: 72,
          symbolColumn: 72,
          horizontalDataRegion: 4,
          verticalDataRegion: 4,
          dataCodeWords: 368,
          correctionCodeWords: 144,
          interLeavedBlock: 4,
          interLeavedDataBlock: 92),
      _DataMatrixSymbolAttribute(
          symbolRow: 80,
          symbolColumn: 80,
          horizontalDataRegion: 4,
          verticalDataRegion: 4,
          dataCodeWords: 456,
          correctionCodeWords: 192,
          interLeavedBlock: 4,
          interLeavedDataBlock: 114),
      _DataMatrixSymbolAttribute(
          symbolRow: 88,
          symbolColumn: 88,
          horizontalDataRegion: 4,
          verticalDataRegion: 4,
          dataCodeWords: 576,
          correctionCodeWords: 224,
          interLeavedBlock: 4,
          interLeavedDataBlock: 144),
      _DataMatrixSymbolAttribute(
          symbolRow: 96,
          symbolColumn: 96,
          horizontalDataRegion: 4,
          verticalDataRegion: 4,
          dataCodeWords: 696,
          correctionCodeWords: 272,
          interLeavedBlock: 4,
          interLeavedDataBlock: 174),
      _DataMatrixSymbolAttribute(
          symbolRow: 104,
          symbolColumn: 104,
          horizontalDataRegion: 4,
          verticalDataRegion: 4,
          dataCodeWords: 816,
          correctionCodeWords: 336,
          interLeavedBlock: 6,
          interLeavedDataBlock: 136),
      _DataMatrixSymbolAttribute(
          symbolRow: 120,
          symbolColumn: 120,
          horizontalDataRegion: 6,
          verticalDataRegion: 6,
          dataCodeWords: 1050,
          correctionCodeWords: 408,
          interLeavedBlock: 6,
          interLeavedDataBlock: 175),
      _DataMatrixSymbolAttribute(
          symbolRow: 132,
          symbolColumn: 132,
          horizontalDataRegion: 6,
          verticalDataRegion: 6,
          dataCodeWords: 1304,
          correctionCodeWords: 496,
          interLeavedBlock: 8,
          interLeavedDataBlock: 163),
      _DataMatrixSymbolAttribute(
          symbolRow: 144,
          symbolColumn: 144,
          horizontalDataRegion: 6,
          verticalDataRegion: 6,
          dataCodeWords: 1558,
          correctionCodeWords: 620,
          interLeavedBlock: 10,
          interLeavedDataBlock: 156),

      // Rectangle matrix
      _DataMatrixSymbolAttribute(
          symbolRow: 8,
          symbolColumn: 18,
          horizontalDataRegion: 1,
          verticalDataRegion: 1,
          dataCodeWords: 5,
          correctionCodeWords: 7,
          interLeavedBlock: 1,
          interLeavedDataBlock: 5),
      _DataMatrixSymbolAttribute(
          symbolRow: 8,
          symbolColumn: 32,
          horizontalDataRegion: 2,
          verticalDataRegion: 1,
          dataCodeWords: 10,
          correctionCodeWords: 11,
          interLeavedBlock: 1,
          interLeavedDataBlock: 10),
      _DataMatrixSymbolAttribute(
          symbolRow: 12,
          symbolColumn: 26,
          horizontalDataRegion: 1,
          verticalDataRegion: 1,
          dataCodeWords: 16,
          correctionCodeWords: 14,
          interLeavedBlock: 1,
          interLeavedDataBlock: 16),
      _DataMatrixSymbolAttribute(
          symbolRow: 12,
          symbolColumn: 36,
          horizontalDataRegion: 2,
          verticalDataRegion: 1,
          dataCodeWords: 22,
          correctionCodeWords: 18,
          interLeavedBlock: 1,
          interLeavedDataBlock: 22),
      _DataMatrixSymbolAttribute(
          symbolRow: 16,
          symbolColumn: 36,
          horizontalDataRegion: 2,
          verticalDataRegion: 1,
          dataCodeWords: 32,
          correctionCodeWords: 24,
          interLeavedBlock: 1,
          interLeavedDataBlock: 32),
      _DataMatrixSymbolAttribute(
          symbolRow: 16,
          symbolColumn: 48,
          horizontalDataRegion: 2,
          verticalDataRegion: 1,
          dataCodeWords: 49,
          correctionCodeWords: 28,
          interLeavedBlock: 1,
          interLeavedDataBlock: 49)
    ];

    _creatLogList();
  }

  ///  Method to create the log list
  void _creatLogList() {
    _log = List<int>(256);
    _aLog = List<int>(256);
    _log[0] = -255;
    _aLog[0] = 1;

    for (int i = 1; i <= 255; i++) {
      _aLog[i] = _aLog[i - 1] * 2;

      if (_aLog[i] >= 256) {
        _aLog[i] = _aLog[i] ^ 301;
      }

      _log[_aLog[i]] = i;
    }
  }

  /// method to encode the provided data
  List<int> _getData() {
    return List<int>.from(utf8.encode(_inputValue));
  }

  /// Method used to create the data matrix
  void _createMatrix(List<int> _codeword) {
    int _x, _y, _numOfRows, _numOfColumns;
    List<int> _places;
    final int _width = _symbolAttribute.symbolColumn;
    final int _height = _symbolAttribute.symbolRow;
    final int _fieldWidth = _width ~/ _symbolAttribute.horizontalDataRegion;
    final int _fieldHeight = _height ~/ _symbolAttribute.verticalDataRegion;
    _numOfColumns = _width - 2 * (_width ~/ _fieldWidth);
    _numOfRows = _height - 2 * (_height ~/ _fieldHeight);
    _places = List<int>(_numOfColumns * _numOfRows);

    _errorCorrectingCode200Placement(_places, _numOfRows, _numOfColumns);
    final List<int> _matrix = List<int>(_width * _height);
    for (_y = 0; _y < _height; _y += _fieldHeight) {
      for (_x = 0; _x < _width; _x++) {
        _matrix[_y * _width + _x] = 1;
      }

      for (_x = 0; _x < _width; _x += 2) {
        _matrix[(_y + _fieldHeight - 1) * _width + _x] = 1;
      }
    }

    for (_x = 0; _x < _width; _x += _fieldWidth) {
      for (_y = 0; _y < _height; _y++) {
        _matrix[_y * _width + _x] = 1;
      }

      for (_y = 0; _y < _height; _y += 2) {
        _matrix[_y * _width + _x + _fieldWidth - 1] = 1;
      }
    }

    for (_y = 0; _y < _numOfRows; _y++) {
      for (_x = 0; _x < _numOfColumns; _x++) {
        final int _v = _places[(_numOfRows - _y - 1) * _numOfColumns + _x];
        if (_v == 1 ||
            _v > 7 && (_codeword[(_v >> 3) - 1] & (1 << (_v & 7))) != 0) {
          _matrix[(1 + _y + 2 * (_y ~/ (_fieldHeight - 2))) * _width +
              1 +
              _x +
              2 * (_x ~/ (_fieldWidth - 2))] = 1;
        }
      }
    }

    final int _symbolColumn = _symbolAttribute.symbolColumn,
        _symbolRow = _symbolAttribute.symbolRow;

    final List<List<int>> _tempArray = List<List<int>>.generate(
        _symbolColumn, (int j) => List<int>(_symbolRow));

    for (int _m = 0; _m < _symbolColumn; _m++) {
      for (int _n = 0; _n < _symbolRow; _n++) {
        _tempArray[_m][_n] = _matrix[_symbolColumn * _n + _m];
      }
    }

    final List<List<int>> _tempArray2 = List<List<int>>.generate(
        _symbolRow, (int j) => List<int>(_symbolColumn));

    for (int i = 0; i < _symbolRow; i++) {
      for (int j = 0; j < _symbolColumn; j++) {
        _tempArray2[_symbolRow - 1 - i][j] = _tempArray[j][i];
      }
    }

    _addQuietZone(_tempArray2);
  }

  /// Method used to add the quiet zone
  void _addQuietZone(List<List<int>> _tempArray) {
    const int _quietZone = 1;
    final int _w = _symbolAttribute.symbolRow + (2 * _quietZone);
    final int _h = _symbolAttribute.symbolColumn + (2 * _quietZone);
    _dataMatrixList = List<List<int>>.generate(_w, (int j) => List<int>(_h));
    // Top quietzone.
    for (int i = 0; i < _h; i++) {
      _dataMatrixList[0][i] = 0;
    }

    for (int i = _quietZone; i < _w - _quietZone; i++) {
      _dataMatrixList[i][0] = 0;

      for (int j = _quietZone; j < _h - _quietZone; j++) {
        _dataMatrixList[i][j] = _tempArray[i - _quietZone][j - _quietZone];
      }

      // Right quietzone.
      _dataMatrixList[i][_h - _quietZone] = 0;
    }

    for (int i = 0; i < _h; i++) {
      _dataMatrixList[_w - _quietZone][i] = 0;
    }
  }

  /// Method to encode the error correcting code word
  void _errorCorrectingCode200PlacementBit(List<int> _array, int _numOfRows,
      int _numOfColumns, int _row, int _column, int _place, String _character) {
    if (_row < 0) {
      _row += _numOfRows;
      _column += 4 - ((_numOfRows + 4) % 8);
    }

    if (_column < 0) {
      _column += _numOfColumns;
      _row += 4 - ((_numOfColumns + 4) % 8);
    }

    _array[_row * _numOfColumns + _column] =
        (_place << 3) + _character.codeUnitAt(0);
  }

  /// Method to encode the error correcting code word
  void _errorCorrectingCode200PlacementCornerA(
      List<int> _array, int _numOfRows, int _numOfColumns, int _place) {
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns,
        _numOfRows - 1, 0, _place, String.fromCharCode(7));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns,
        _numOfRows - 1, 1, _place, String.fromCharCode(6));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns,
        _numOfRows - 1, 2, _place, String.fromCharCode(5));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 0,
        _numOfColumns - 2, _place, String.fromCharCode(4));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 0,
        _numOfColumns - 1, _place, String.fromCharCode(3));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 1,
        _numOfColumns - 1, _place, String.fromCharCode(2));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 2,
        _numOfColumns - 1, _place, String.fromCharCode(1));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 3,
        _numOfColumns - 1, _place, String.fromCharCode(0));
  }

  /// Method to encode the error correcting code word
  void _errorCorrectingCode200PlacementCornerB(
      List<int> _array, int _numOfRows, int _numOfColumns, int _place) {
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns,
        _numOfRows - 3, 0, _place, String.fromCharCode(7));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns,
        _numOfRows - 2, 0, _place, String.fromCharCode(6));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns,
        _numOfRows - 1, 0, _place, String.fromCharCode(5));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 0,
        _numOfColumns - 4, _place, String.fromCharCode(4));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 0,
        _numOfColumns - 3, _place, String.fromCharCode(3));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 0,
        _numOfColumns - 2, _place, String.fromCharCode(2));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 0,
        _numOfColumns - 1, _place, String.fromCharCode(1));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 1,
        _numOfColumns - 1, _place, String.fromCharCode(0));
  }

  /// Method to encode the error correcting code word
  void _errorCorrectingCode200PlacementCornerC(
      List<int> _array, int _numOfRows, int _numOfColumns, int _place) {
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns,
        _numOfRows - 3, 0, _place, String.fromCharCode(7));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns,
        _numOfRows - 2, 0, _place, String.fromCharCode(6));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns,
        _numOfRows - 1, 0, _place, String.fromCharCode(5));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 0,
        _numOfColumns - 2, _place, String.fromCharCode(4));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 0,
        _numOfColumns - 1, _place, String.fromCharCode(3));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 1,
        _numOfColumns - 1, _place, String.fromCharCode(2));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 2,
        _numOfColumns - 1, _place, String.fromCharCode(1));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 3,
        _numOfColumns - 1, _place, String.fromCharCode(0));
  }

  /// Method to encode the error correcting code word
  void _errorCorrectingCode200PlacementCornerD(
      List<int> _array, int _numOfRows, int _numOfColumns, int _place) {
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns,
        _numOfRows - 1, 0, _place, String.fromCharCode(7));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns,
        _numOfRows - 1, _numOfColumns - 1, _place, String.fromCharCode(6));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 0,
        _numOfColumns - 3, _place, String.fromCharCode(5));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 0,
        _numOfColumns - 2, _place, String.fromCharCode(4));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 0,
        _numOfColumns - 1, _place, String.fromCharCode(3));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 1,
        _numOfColumns - 3, _place, String.fromCharCode(2));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 1,
        _numOfColumns - 2, _place, String.fromCharCode(1));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, 1,
        _numOfColumns - 1, _place, String.fromCharCode(0));
  }

  /// Method to encode the error correcting code word
  void _errorCorrectingCode200PlacementBlock(List<int> _array, int _numOfRows,
      int _numOfColumns, int _row, int _column, int _place) {
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns,
        _row - 2, _column - 2, _place, String.fromCharCode(7));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns,
        _row - 2, _column - 1, _place, String.fromCharCode(6));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns,
        _row - 1, _column - 2, _place, String.fromCharCode(5));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns,
        _row - 1, _column - 1, _place, String.fromCharCode(4));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns,
        _row - 1, _column, _place, String.fromCharCode(3));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, _row,
        _column - 2, _place, String.fromCharCode(2));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, _row,
        _column - 1, _place, String.fromCharCode(1));
    _errorCorrectingCode200PlacementBit(_array, _numOfRows, _numOfColumns, _row,
        _column, _place, String.fromCharCode(0));
  }

  /// Method to encode the error correcting code word
  void _errorCorrectingCode200Placement(
      List<int> _array, int _numOfRows, int _numOfColumns) {
    int _row, _column, _place;
    for (_row = 0; _row < _numOfRows; _row++) {
      for (_column = 0; _column < _numOfColumns; _column++) {
        _array[_row * _numOfColumns + _column] = 0;
      }
    }

    _place = 1;
    _row = 4;
    _column = 0;
    do {
      if (_row == _numOfRows && !(_column != 0)) {
        _errorCorrectingCode200PlacementCornerA(
            _array, _numOfRows, _numOfColumns, _place++);
      }

      if ((_row == _numOfRows - 2) &&
          !(_column != 0) &&
          ((_numOfColumns % 4) != 0)) {
        _errorCorrectingCode200PlacementCornerB(
            _array, _numOfRows, _numOfColumns, _place++);
      }

      if (_row == _numOfRows - 2 &&
          !(_column != 0) &&
          (_numOfColumns % 8) == 4) {
        _errorCorrectingCode200PlacementCornerC(
            _array, _numOfRows, _numOfColumns, _place++);
      }

      if (_row == _numOfRows + 4 &&
          _column == 2 &&
          !((_numOfColumns % 8) != 0)) {
        _errorCorrectingCode200PlacementCornerD(
            _array, _numOfRows, _numOfColumns, _place++);
      }
      // enocoding placement (up/right)
      do {
        if (_row < _numOfRows &&
            _column >= 0 &&
            !(_array[_row * _numOfColumns + _column] != 0)) {
          _errorCorrectingCode200PlacementBlock(
              _array, _numOfRows, _numOfColumns, _row, _column, _place++);
        }

        _row -= 2;
        _column += 2;
      } while (_row >= 0 && _column < _numOfColumns);
      _row++;
      _column += 3;
      // enocoding placement (down/left)
      do {
        if (_row >= 0 &&
            _column < _numOfColumns &&
            !(_array[_row * _numOfColumns + _column] != 0)) {
          _errorCorrectingCode200PlacementBlock(
              _array, _numOfRows, _numOfColumns, _row, _column, _place++);
        }

        _row += 2;
        _column -= 2;
      } while (_row < _numOfRows && _column >= 0);
      _row += 3;
      _column++;
    } while (_row < _numOfRows || _column < _numOfColumns);
    if (!(_array[_numOfRows * _numOfColumns - 1] != 0)) {
      _array[_numOfRows * _numOfColumns - 1] =
          _array[_numOfRows * _numOfColumns - _numOfColumns - 2] = 1;
    }
  }

  /// Method to double the error correcting code
  int _errorCorrectingCodeDoublify(int i, int j) {
    if (i == 0) {
      return 0;
    }

    if (j == 0) {
      return i;
    }

    return _aLog[(_log[i] + j) % 255];
  }

  /// Method used for error correction
  int _errorCorrectingCodeProduct(int i, int j) {
    if (i == 0 || j == 0) {
      return 0;
    } else if (i > 255 || j > 255) {
      return 0;
    }
    return _aLog[(_log[i] + _log[j]) % 255];
  }

  /// Method to perform the XOR operation for error correction
  int _errorCorrectingCodeSum(int i, int j) {
    return i ^ j;
  }

  /// Method to pad the code word
  List<int> _padCodewords(int _dataCWLength, List<int> _temp) {
    final List<int> _codewords = <int>[];
    int _length = _temp.length;
    for (int i = 0; i < _length; i++) {
      _codewords.add(_temp[i]);
    }

    if (_length < _dataCWLength) {
      _codewords.add(129);
    }

    _length = _codewords.length;
    while (_length < _dataCWLength) {
      int _value = 129 + (((_length + 1) * 149) % 253) + 1;
      if (_value > 254) {
        _value -= 254;
      }

      _codewords.add(_value);
      _length = _codewords.length;
    }

    return _codewords;
  }

  /// Method used for base256 encoding
  List<int> _dataMatrixBaseEncoder(List<int> _dataCodeword) {
    int _num = 1;
    if (_dataCodeword.length > 249) {
      _num++;
    }

    final List<int> _result = List<int>((1 + _num) + _dataCodeword.length);
    _result[0] = 231;
    if (_dataCodeword.length <= 249) {
      _result[1] = _dataCodeword.length;
    } else {
      _result[1] = ((_dataCodeword.length / 250) + 249).toInt();
      _result[2] = _dataCodeword.length % 250;
    }

    for (int i = 0; i < _dataCodeword.length; i++) {
      _result[(1 + _num) + i] = _dataCodeword[i];
    }

    for (int i = 1; i < _result.length; i++) {
      _result[i] = _computeBase256Codeword(_result[i], i);
    }

    return _result;
  }

  /// Method to compute the code word
  int _computeBase256Codeword(int _codewordValue, int _index) {
    final int i = ((149 * (_index + 1)) % 255) + 1;
    final int j = _codewordValue + i;
    if (j <= 255) {
      return j;
    }

    return j - 256;
  }

  /// Method used for ASCII numeric encoding
  List<int> _dataMatrixASCIINumericEncoder(List<int> _dataCodeword) {
    List<int> _destinationArray = _dataCodeword;
    bool _isEven = true;
    // if the input data length is odd, add 0 in front of the data.
    if ((_destinationArray.length % 2) == 1) {
      _isEven = false;
      _destinationArray = List<int>(_dataCodeword.length + 1);
      for (int i = 0; i < _dataCodeword.length; i++) {
        _destinationArray[i] = _dataCodeword[i];
      }
    }

    final List<int> _result = List<int>(_destinationArray.length ~/ 2);
    for (int i = 0; i < _result.length; i++) {
      if (!_isEven && i == _result.length - 1) {
        _result[i] = _destinationArray[2 * i] + 1;
      } else {
        _result[i] = (((_destinationArray[2 * i] - 48) * 10) +
                (_destinationArray[(2 * i) + 1] - 48)) +
            130;
      }
    }
    return _result;
  }

  /// Method used for ASCII encoding
  List<int> _dataMatrixASCIIEncoder(List<int> _dataCodeword) {
    final List<int> _result = List<int>.from(_dataCodeword);
    int _index = 0;
    for (int i = 0; i < _dataCodeword.length; i++) {
      if (_dataCodeword[i] >= 48 && _dataCodeword[i] <= 57) {
        int _prevIndex = 0;
        if (i != 0) {
          _prevIndex = _index - 1;
        }

        final int _prevValue = _result[_prevIndex] - 1;
        int _priorValue = 0;
        if (i != 0 && _index != 1) {
          _priorValue = _result[_prevIndex - 1];
        }

        //Check the priorvalue is digit or non convertable value.if it is true then combine the 2 digit
        if (_priorValue != 235 && _prevValue >= 48 && _prevValue <= 57) {
          _result[_prevIndex] =
              10 * (_prevValue - 0) + (_dataCodeword[i] - 0) + 130;
        } else {
          _result[_index++] = _dataCodeword[i] + 1;
        }
      } else if (_dataCodeword[i] < 127) {
        _result[_index++] = _dataCodeword[i] + 1;
      } else {
        //Assign 235 default value for non
        // convertable values(other than digits,letters and special characters 0 to 127 asciii values)
        _result[_index] = 235;
        _result[_index++] = _dataCodeword[i] - 127;
      }
    }

    final List<int> _encodedData = List<int>.from(_result);
    return _encodedData;
  }

  /// Method used for computing error correction
  List<int> _computeErrorCorrection(List<int> _codeword) {
    int _dataLength = _codeword.length;
    _symbolAttribute = _DataMatrixSymbolAttribute();
    if (dataMatrixSize == DataMatrixSize.auto) {
      for (int i = 0; i < _symbolAttributes.length; i++) {
        final _DataMatrixSymbolAttribute _attribute = _symbolAttributes[i];
        if (_attribute.dataCodeWords >= _dataLength) {
          _symbolAttribute = _attribute;
          break;
        }
      }
    } else {
      _symbolAttribute =
          _symbolAttributes[_getDataMatrixSize(dataMatrixSize) - 1];
    }

    List<int> _temp;
    // Pad data codeword if the length is less than the selected symbol attribute.
    if (_symbolAttribute.dataCodeWords > _dataLength) {
      _temp = _padCodewords(_symbolAttribute.dataCodeWords, _codeword);
      _encodedCodeword = List<int>.from(_temp);
      _dataLength = _codeword.length;
    } else if (_symbolAttribute.dataCodeWords == 0) {
      throw 'Data cannot be encoded as barcode';
    } else if (_symbolAttribute.dataCodeWords < _dataLength) {
      final String _symbolRow = _symbolAttribute.symbolRow.toString();
      final String _symbolColumn = _symbolAttribute.symbolColumn.toString();
      throw 'Data too long for $_symbolRow x $_symbolColumn barcode.';
    }

    final int _numCorrectionCodeword = _symbolAttribute.correctionCodeWords;

    // Create error correction array.

    List<int> _correctionCodeWordArray =
        List<int>(_numCorrectionCodeword + _symbolAttribute.dataCodeWords);
    for (int i = 0; i < _correctionCodeWordArray.length; i++) {
      _correctionCodeWordArray[i] = 0;
    }
    final int _step = _symbolAttribute.interLeavedBlock;
    final int _symbolDataWords = _symbolAttribute.dataCodeWords;
    final int _blockErrorWords = _symbolAttribute.correctionCodeWords ~/ _step;
    final int _total = _symbolDataWords + _blockErrorWords * _step;

    // Updates m_polynomial based on the required number of correction bytes.

    _createPolynomial(_step);
    //set block length (each block consists of length 68 )
    _blockLength = 68;
    final List<int> _blockByte = List<int>(_blockLength);
    for (int _block = 0; _block < _step; _block++) {
      for (int k = 0; k < _blockByte.length; k++) {
        _blockByte[k] = 0;
      }

      for (int i = _block; i < _symbolDataWords; i += _step) {
        final int _val = _errorCorrectingCodeSum(
            _blockByte[_blockErrorWords - 1], _encodedCodeword[i]);
        for (int j = _blockErrorWords - 1; j > 0; j--) {
          _blockByte[j] = _errorCorrectingCodeSum(_blockByte[j - 1],
              _errorCorrectingCodeProduct(_polynomial[j], _val));
        }

        _blockByte[0] = _errorCorrectingCodeProduct(_polynomial[0], _val);
      }

      int _blockDataWords = 0;
      if (_block >= 8 && dataMatrixSize == DataMatrixSize.size144x144) {
        _blockDataWords = _symbolAttribute.dataCodeWords ~/ _step;
      } else {
        _blockDataWords = _symbolAttribute.interLeavedDataBlock;

        int _bIndex = _blockErrorWords;

        for (int i = _block + (_step * _blockDataWords);
            i < _total;
            i += _step) {
          _correctionCodeWordArray[i] = _blockByte[--_bIndex];
        }

        if (_bIndex != 0) {
          throw 'Error in error correction code generation!';
        }
      }
    }

    if (_correctionCodeWordArray.length > _numCorrectionCodeword) {
      final List<int> _tmp = _correctionCodeWordArray;
      _correctionCodeWordArray = List<int>(_numCorrectionCodeword);
      for (int i = 0; i < _correctionCodeWordArray.length; i++) {
        _correctionCodeWordArray[i] = 0;
      }
      int z = 0;

      for (int i = _tmp.length - 1; i > _symbolAttribute.dataCodeWords; i--) {
        _correctionCodeWordArray[z++] = _tmp[i];
      }
    }

    final List<int> _reversedList = _correctionCodeWordArray.reversed.toList();
    return _reversedList;
  }

  /// Method used to create the polynomial
  void _createPolynomial(int _step) {
    //Set block length for polynomial values
    _blockLength = 69;
    _polynomial = List<int>(_blockLength);
    final int _blockErrorWords = _symbolAttribute.correctionCodeWords ~/ _step;
    for (int i = 0; i < _polynomial.length; i++) {
      _polynomial[i] = 0x01;
    }

    for (int i = 1; i <= _blockErrorWords; i++) {
      for (int j = i - 1; j >= 0; j--) {
        _polynomial[j] = _errorCorrectingCodeDoublify(_polynomial[j], i);
        if (j > 0) {
          _polynomial[j] =
              _errorCorrectingCodeSum(_polynomial[j], _polynomial[j - 1]);
        }
      }
    }
  }

  List<int> _prepareCodeword(List<int> _dataCodeword) {
    _encodedCodeword = _prepareDataCodeword(_dataCodeword);
    final List<int> _correctCodeword =
        _computeErrorCorrection(_encodedCodeword);
    final List<int> _finalCodeword =
        List<int>(_encodedCodeword.length + _correctCodeword.length);
    for (int i = 0; i < _encodedCodeword.length; i++) {
      _finalCodeword[i] = _encodedCodeword[i];
    }

    for (int i = 0; i < _correctCodeword.length; i++) {
      _finalCodeword[i + _encodedCodeword.length] = _correctCodeword[i];
    }

    return _finalCodeword;
  }

  /// Method used to prepare the code word
  List<int> _prepareDataCodeword(List<int> _dataCodeword) {
    _encoding = encoding;
    if (encoding == DataMatrixEncoding.auto ||
        encoding == DataMatrixEncoding.asciiNumeric) {
      bool _number = true;
      bool _extended = false;
      int _num = 0;
      final List<int> _data = _dataCodeword;
      DataMatrixEncoding _actualEncoding = DataMatrixEncoding.ascii;
      for (int i = 0; i < _data.length; i++) {
        if ((_data[i] < 48) || (_data[i] > 57)) {
          _number = false;
        } else if (_data[i] > 127) {
          _num++;
          if (_num > 3) {
            _extended = true;
            break;
          }
        }
      }

      if (_number) {
        _actualEncoding = DataMatrixEncoding.asciiNumeric;
      }

      if (_extended) {
        _actualEncoding = DataMatrixEncoding.base256;
      }

      if (_actualEncoding == DataMatrixEncoding.asciiNumeric &&
          encoding != _actualEncoding) {
        throw 'Data contains invalid characters and cannot be encoded as ASCIINumeric.';
      }

      _encoding = _actualEncoding;
    }

    List<int> _result;
    switch (_encoding) {
      case DataMatrixEncoding.ascii:
        _result = _dataMatrixASCIIEncoder(_dataCodeword);
        break;
      case DataMatrixEncoding.asciiNumeric:
        _result = _dataMatrixASCIINumericEncoder(_dataCodeword);
        break;
      case DataMatrixEncoding.base256:
        _result = _dataMatrixBaseEncoder(_dataCodeword);
        break;
      case DataMatrixEncoding.auto:
        break;
    }

    return _result;
  }

  @override
  bool _validateInput(String value) {
    return true;
  }

  @override
  void _renderBarcode(
      Canvas canvas,
      Size size,
      Offset offset,
      String value,
      Color foregroundColor,
      TextStyle textStyle,
      double textSpacing,
      TextAlign textAlign,
      bool showValue) {
    _inputValue = value;

    _buildDataMatrix();
    const int _quietZone = 1;
    double x = 0;
    double y = 0;
    final Paint _paint = _getBarPaint(foregroundColor);
    final int _w = _symbolAttribute.symbolRow + (2 * _quietZone);
    final int _h = _symbolAttribute.symbolColumn + (2 * _quietZone);
    final double _minSize =
        size.width >= size.height ? size.height : size.width;
    final bool _isSquareMatrix = _getDataMatrixSize(dataMatrixSize) < 25;
    int _dimension = _minSize ~/ _dataMatrixList.length;
    final int _rectDimension = _minSize ~/ _dataMatrixList[0].length;
    final int _xDimension =
        module != null ? module : _isSquareMatrix ? _dimension : _rectDimension;
    _dimension = module != null ? module : _dimension;
    for (int i = 0; i < _w; i++) {
      x = 0;
      for (int j = 0; j < _h; j++) {
        x += _xDimension;
      }

      y += _dimension;
    }

    final double _xPosition = offset.dx + (size.width - x) / 2;
    double _yPosition = offset.dy + (size.height - y) / 2;

    for (int i = 0; i < _w; i++) {
      x = _xPosition;
      for (int j = 0; j < _h; j++) {
        if (_dataMatrixList[i][j] == 1) {
          final Rect _matrixRect = Rect.fromLTRB(
              x, _yPosition, x + _xDimension, _yPosition + _dimension);
          canvas.drawRect(_matrixRect, _paint);
        }

        x += _xDimension;
      }

      _yPosition += _dimension;
    }

    if (showValue) {
      final Offset _offset = Offset(offset.dx, _yPosition.toDouble());
      _drawText(
          canvas, _offset, size, value, textStyle, textSpacing, textAlign);
    }
  }

  /// Method to render the input value of the barcode
  void _drawText(Canvas canvas, Offset offset, Size size, String value,
      TextStyle textStyle, double textSpacing, TextAlign textAlign,
      [Offset _actualOffset, Size _actualSize]) {
    final TextSpan _span = TextSpan(text: value, style: textStyle);

    final TextPainter _textPainter = TextPainter(
        maxLines: 1,
        ellipsis: '.....',
        text: _span,
        textDirection: TextDirection.ltr,
        textAlign: textAlign);
    _textPainter.layout(minWidth: 0, maxWidth: size.width);
    double _x;
    double _y;
    switch (textAlign) {
      case TextAlign.justify:
      case TextAlign.center:
        {
          _x = offset.dx + (size.width / 2 - _textPainter.width / 2);
          _y = offset.dy + textSpacing;
        }
        break;
      case TextAlign.left:
      case TextAlign.start:
        {
          _x = offset.dx;
          _y = offset.dy + textSpacing;
        }
        break;
      case TextAlign.right:
      case TextAlign.end:
        {
          _x = offset.dx + (size.width - _textPainter.width);
          _y = offset.dy + textSpacing;
        }
        break;
    }
    _textPainter.paint(canvas, Offset(_x, _y));
  }

  void _buildDataMatrix() {
    final List<int> _codeword = _prepareCodeword(_getData());
    _createMatrix(_codeword);
  }
}

/// Represents the data matrix symbol attribute
class _DataMatrixSymbolAttribute {
  /// Creates the data matrix symbol attribute
  _DataMatrixSymbolAttribute(
      {this.symbolRow,
      this.symbolColumn,
      this.horizontalDataRegion,
      this.verticalDataRegion,
      this.dataCodeWords,
      this.correctionCodeWords,
      this.interLeavedBlock,
      this.interLeavedDataBlock});

  /// Defines the symbol row
  final int symbolRow;

  /// Defines the symbol column
  final int symbolColumn;

  /// Defines the horizontal data region
  final int horizontalDataRegion;

  /// Defines the vertical data region
  final int verticalDataRegion;

  /// Defines the data code words
  final int dataCodeWords;

  /// Defines the error correction code words
  final int correctionCodeWords;

  /// Defines the inter leaved blocks
  final int interLeavedBlock;

  /// Defines the inter leaved data blocks
  final int interLeavedDataBlock;
}
