part of barcodes;

/// Quick response code ([QRCode]) is a two-dimensional barcode. It can efficiently store more information
/// in a smaller space than 1D barcodes. Each barcode can store values up to 7089 characters.
/// It is mostly used for URLs, business cards, contact information, etc.
///
/// The [QRCode] consists of a grid of dark and light dots or blocks that form a square.
/// The data encoded in the barcode can be numeric, alphanumeric, or Shift JIS characters.
///
/// * The QR Code uses version from 1 to 40. Version 1 measures 21 modules x 21 modules, version 2 measures 25 modules x 25 modules, and so on.
/// The number of modules increases in steps of 4 modules per side up to version 40 that measures 177 modules x 177 modules.
/// * Each version has its own capacity. By default, the barcode control automatically sets the version according to the length of the input text.
/// * The [QRCode] is designed for industrial uses and also commonly used in consumer advertising.
///
/// The amount of data that can be stored in the [QRCode] symbol depends on the [inputMode], [codeVersion] (1, ..., 40, indicating the overall dimensions of the symbol, i.e. 4 × version number + 17 dots on each side),
/// and [errorCorrectionLevel]. The maximum storage capacities occur for [codeVersion] 40 and [ErrorCorrectionLevel.low].
class QRCode extends Symbology {
  /// Create a [QRCode] symbology with the default or required properties.
  ///
  /// The arguments [module] must be non-negative and greater than 0.
  ///
  QRCode(
      {this.codeVersion,
      this.errorCorrectionLevel = ErrorCorrectionLevel.high,
      this.inputMode = QRInputMode.binary,
      int module})
      : super(module: module) {
    if (codeVersion != null && codeVersion != QRCodeVersion.auto) {
      _codeVersion = codeVersion;
      _isUserMentionedVersion = true;
    } else {
      _isUserMentionedVersion = false;
    }

    if (errorCorrectionLevel != null) {
      _errorCorrectionLevel = errorCorrectionLevel;
      _isUserMentionedErrorCorrectionLevel = true;
    } else {
      _isUserMentionedErrorCorrectionLevel = false;
    }

    if (inputMode != null) {
      _inputMode = inputMode;
      _isUserMentionedMode = true;
    } else {
      _isUserMentionedMode = false;
    }
  }

  /// Define the version that is used to encode the amount of data.
  ///
  /// The [QRCode] uses version from 1 to 40. Version 1 measures 21 modules x 21 modules, version 2 measures 25 modules x 25 modules, and so on.
  ///
  /// The number of modules increases in steps of 4 modules per side up to version 40 that measures 177 modules x 177 modules.
  ///
  /// Defaults to [QRCodeVersion.auto].
  ///
  /// Also refer [QRCodeVersion].
  ///
  /// ```dart
  /// Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfBarcodeGenerator(value:'123456',
  ///        symbology: QRCode(codeVersion: 4)));
  ///}
  /// ```dart
  final QRCodeVersion codeVersion;

  /// Define the encode recovery capacity of the barcode.
  ///
  /// Below error correction capability at each of the four levels:
  ///
  /// * [ErrorCorrectionLevel.low] - approximately 7% of the codewords can be restored.
  ///
  /// * [ErrorCorrectionLevel.medium] - approximately 15% of the codewords can be restored.
  ///
  /// * [ErrorCorrectionLevel.quartile] - approximately 25% of the codewords can be restored.
  ///
  /// * [ErrorCorrectionLevel.high] - approximately 30% of the codewords can be restored.
  ///
  /// Defaults to [ErrorCorrectionLevel.high].
  ///
  /// Also refer [ErrorCorrectionLevel].
  ///
  /// ```dart
  /// Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfBarcodeGenerator(value:'123456',
  ///        symbology: QRCode(errorCorrectionLevel: ErrorCorrectionLevel.high)));
  ///}
  /// ```dart
  final ErrorCorrectionLevel errorCorrectionLevel;

  /// Define a specific set of input mode characters.
  ///
  /// Defaults to [QRInputMode.binary].
  ///
  ///  Also refer [QRInputMode].
  ///
  /// ```dart
  /// Widget build(BuildContext context) {
  ///    return Container(
  ///        child: SfBarcodeGenerator(value:'123456',
  ///        symbology: QRCode(inputMode: m QRInputMode.low));
  ///}
  /// ```dart
  final QRInputMode inputMode;

  /// Specifies the CP437 character set
  static const List<String> _cp437CharSet = <String>[
    '2591',
    '2592',
    '2593',
    '2502',
    '2524',
    '2561',
    '2562',
    '2556',
    '2555',
    '2563',
    '2551',
    '2557',
    '255D',
    '255C',
    '255B',
    '2510',
    '2514',
    '2534',
    '252C',
    '251C',
    '2500',
    '253C',
    '255E',
    '255F',
    '255A',
    '2554',
    '2569',
    '2566',
    '2560',
    '2550',
    '256C',
    '2567',
    '2568',
    '2564',
    '2565',
    '2559',
    '2558',
    '2552',
    '2553',
    '256B',
    '256A',
    '2518',
    '250C',
    '2588',
    '2584',
    '258C',
    '2590',
    '2580',
    '25A0'
  ];

  /// Specifies the latin2 character set
  static const List<String> _latin2CharSet = <String>[
    '104',
    '2D8',
    '141',
    '13D',
    '15A',
    '160',
    '15E',
    '164',
    '179',
    '17D',
    '17B',
    '105',
    '2DB',
    '142',
    '13E',
    '15B',
    '2C7',
    '161',
    '15F',
    '165',
    '17A',
    '2DD',
    '17E',
    '17C',
    '154',
    '102',
    '139',
    '106',
    '10C',
    '118',
    '11A',
    '10E',
    '110',
    '143',
    '147',
    '150',
    '158',
    '16E',
    '170',
    '162',
    '155',
    '103',
    '13A',
    '107',
    '10D',
    '119',
    '11B',
    '10F',
    '111',
    '144',
    '148',
    '151',
    '159',
    '16F',
    '171',
    '163',
    '2D9'
  ];

  /// Specifies the latin3 character set
  static const List<String> _latin3CharSet = <String>[
    '126',
    '124',
    '130',
    '15E',
    '11E',
    '134',
    '17B',
    '127',
    '125',
    '131',
    '15F',
    '11F',
    '135',
    '17C',
    '10A',
    '108',
    '120',
    '11C',
    '16C',
    '15C',
    '10B',
    '109',
    '121',
    '11D',
    '16D',
    '15D'
  ];

  /// Specifies the latin4 character set
  static const List<String> _latin4CharSet = <String>[
    '104',
    '138',
    '156',
    '128',
    '13B',
    '160',
    '112',
    '122',
    '166',
    '17D',
    '105',
    '2DB',
    '157',
    '129',
    '13C',
    '2C7',
    '161',
    '113',
    '123',
    '167',
    '14A',
    '17E',
    '14B',
    '100',
    '12E',
    '10C',
    '118',
    '116',
    '12A',
    '110',
    '145',
    '14C',
    '136',
    '172',
    '168',
    '16A',
    '101',
    '12F',
    '10D',
    '119',
    '117',
    '12B',
    '111',
    '146',
    '14D',
    '137',
    '173',
    '169',
    '16B'
  ];

  /// Specifies the windows 1250 character set
  static const List<String> _windows1250CharSet = <String>[
    '141',
    '104',
    '15E',
    '17B',
    '142',
    '105',
    '15F',
    '13D',
    '13E',
    '17C'
  ];

  /// Specifies the windows 1251 character set
  static const List<String> _windows1251CharSet = <String>[
    '402',
    '403',
    '453',
    '409',
    '40A',
    '40C',
    '40B',
    '40F',
    '452',
    '459',
    '45A',
    '45C',
    '45B',
    '45F',
    '40E',
    '45E',
    '408',
    '490',
    '401',
    '404',
    '407',
    '406',
    '456',
    '491',
    '451',
    '454',
    '458',
    '405',
    '455',
    '457'
  ];

  /// Specifies the windows 1252 character set
  static const List<String> _windows1252CharSet = <String>[
    '20AC',
    '201A',
    '192',
    '201E',
    '2026',
    '2020',
    '202',
    '2C6',
    '2030',
    '160',
    '2039',
    '152',
    '17D',
    '2018',
    '2019',
    '201C',
    '201D',
    '2022',
    '2013',
    '2014',
    '2DC',
    '2122',
    '161',
    '203A',
    '153',
    '17E',
    '178'
  ];

  /// Specifies the QR code value
  _QRCodeValue _qrCodeValues;

  /// Specifies the QR code version
  QRCodeVersion _codeVersion = QRCodeVersion.version01;

  /// Species the module value
  int _noOfModules = 21;

  /// Specifies the list of module value
  List<List<_ModuleValue>> _moduleValues;

  /// Specifies the data alllocation value
  List<List<_ModuleValue>> _dataAllocationValues;

  /// Specifies the actual input mode
  QRInputMode _inputMode = QRInputMode.numeric;

  /// Specifies the actula error corrceton level
  ErrorCorrectionLevel _errorCorrectionLevel = ErrorCorrectionLevel.low;

  /// Specifies the dat bits value
  int _dataBits = 0;

  /// Specifies the list of blocks
  List<int> _blocks;

  /// Specifies the user mentioned level
  bool _isUserMentionedMode = false;

  /// Specifies the version is mentioned by user
  bool _isUserMentionedVersion = false;

  /// Specifies the error correction level is mentioned by user
  bool _isUserMentionedErrorCorrectionLevel = false;

  /// Specifies the bool value based on the error correction input
  bool _isEci = false;

  /// Specifies the assignment number
  int _eciAssignmentNumber = 3;

  /// Specifies the text value
  String _encodedText;

  /// Specifies the list of encode data
  List<bool> _encodeDataCodeWords;

  @override
  bool _validateInput(String value) {
    return true;
  }

  /// Specifies the character set of CP437
  bool _isCP437Character(int _inputChar) {
    final String _inputCharIndex = _inputChar.toRadixString(16);
    if (_cp437CharSet.contains(_inputCharIndex)) {
      return true;
    }

    return false;
  }

  /// Gets the input character is present in ISO8859_2 character set
  bool _isISO8859_2CharacterSet(int _input) {
    final String _inputInHex = _input.toRadixString(16);
    if (_latin2CharSet.contains(_inputInHex)) {
      return true;
    }
    return false;
  }

  /// Checks the input character is present in ISO8859_3 character set
  bool _isISO8859_3CharacterSet(int _input) {
    final String _inputInHex = _input.toRadixString(16);

    if (_latin3CharSet.contains(_inputInHex)) {
      return true;
    }
    return false;
  }

  /// Checks the input character is present in ISO8859_4 character set
  bool _isISO8859_4CharacterSet(int _input) {
    final String _inputInHex = _input.toRadixString(16);
    if (_latin4CharSet.contains(_inputInHex)) {
      return true;
    }
    return false;
  }

  /// Checks the input character is present in ISO8859_4 character set
  bool _isISO8859_5CharacterSet(int _input) {
    if (_input >= 1025 &&
        _input <= 1119 &&
        _input != 1037 &&
        _input != 1104 &&
        _input != 1117) {
      return true;
    }

    return false;
  }

  ///Checks the input character is present in ISO8859_5 character set
  bool _isISO8859_6CharacterSet(int _input) {
    if ((_input >= 1569 && _input <= 1594) ||
        (_input >= 1600 && _input <= 1618) ||
        _input == 1567 ||
        _input == 1563 ||
        _input == 1548) {
      return true;
    }

    return false;
  }

  ///Checks the input character is present in ISO8859_7 character set
  bool _isISO8859_7CharacterSet(int _input) {
    if ((_input >= 900 && _input <= 974) || _input == 890) {
      return true;
    }

    return false;
  }

  /// Checks the input character is present in ISO8859_8 character set
  bool _isISO8859_8CharacterSet(int _input) {
    if (_input >= 1488 && _input <= 1514) {
      return true;
    }

    return false;
  }

  /// Checks the input character is present in ISO8859_811character set
  bool _isISO8859_11CharacterSet(int _input) {
    if (_input >= 3585 && _input <= 3675) {
      return true;
    }

    return false;
  }

  /// Checks the input character is present in Windows1250Character  set
  bool _isWindows1250Character(int _input) {
    final String _inputCharHex = _input.toRadixString(16);
    if (_windows1250CharSet.contains(_inputCharHex)) {
      return true;
    }

    if (_input >= 1569 && _input <= 1610) {
      return true;
    }

    return false;
  }

  /// Checks the input character is present in Windows1251Character  set
  bool _isWindows1251Character(int _input) {
    final String _inputChar = _input.toRadixString(16);
    if (_windows1251CharSet.contains(_inputChar)) {
      return true;
    }

    if (_input >= 1040 && _input <= 1103) {
      return true;
    }

    return false;
  }

  /// Checks the input character is present in Windows1252Character  set
  bool _isWindows1252Character(int _input) {
    final String _inputChar = _input.toRadixString(16);

    if (_windows1252CharSet.contains(_inputChar)) {
      return true;
    }

    return false;
  }

  /// Checks the input character is present in Windows1256Character  set
  bool _isWindows1256Character(int _input) {
    final String _inputChar = _input.toRadixString(16);
    final List<String> _windows1256CharSet = <String>[
      '67E',
      '679',
      '152',
      '686',
      '698',
      '688',
      '6AF',
      '6A9',
      '691',
      '153',
      '6BA',
      '6BE',
      '6C1'
    ];

    if (_windows1256CharSet.contains(_inputChar)) {
      return true;
    }

    if (_input >= 1569 && _input <= 1610) {
      return true;
    }

    return false;
  }

  /// Allocates the format and the version information
  void _allocateFormatAndVersionInformation() {
    for (int i = 0; i < 9; i++) {
      _moduleValues[8][i]._isFilled = true;
      _moduleValues[i][8]._isFilled = true;
    }

    for (int i = _noOfModules - 8; i < _noOfModules; i++) {
      _moduleValues[8][i]._isFilled = true;
      _moduleValues[i][8]._isFilled = true;
    }

    final int _version = _getVersionNumber(_codeVersion);
    if (_version > 6) {
      final List<int> _versionInformation = _qrCodeValues._versionInformation;
      int _count = 0;
      for (int i = 0; i < 6; i++) {
        for (int j = 2; j >= 0; j--) {
          _moduleValues[i][_noOfModules - 9 - j]._isBlack =
              _versionInformation != null && _versionInformation[_count] == 1
                  ? true
                  : false;
          _moduleValues[i][_noOfModules - 9 - j]._isFilled = true;

          _moduleValues[_noOfModules - 9 - j][i]._isBlack =
              _versionInformation != null && _versionInformation[_count++] == 1
                  ? true
                  : false;
          _moduleValues[_noOfModules - 9 - j][i]._isFilled = true;
        }
      }
    }
  }

  /// Returns the alignment pattern
  List<int> _getAlignmentPatternCoordinates() {
    List<int> _align;
    final int _versionNumber = _getVersionNumber(_codeVersion);
    switch (_versionNumber) {
      case 02:
        _align = <int>[6, 18];
        break;
      case 03:
        _align = <int>[6, 22];
        break;
      case 04:
        _align = <int>[6, 26];
        break;
      case 05:
        _align = <int>[6, 30];
        break;
      case 06:
        _align = <int>[6, 34];
        break;
      case 07:
        _align = <int>[6, 22, 38];
        break;
      case 08:
        _align = <int>[6, 24, 42];
        break;
      case 09:
        _align = <int>[6, 26, 46];
        break;
      case 10:
        _align = <int>[6, 28, 50];
        break;
      case 11:
        _align = <int>[6, 30, 54];
        break;
      case 12:
        _align = <int>[6, 32, 58];
        break;
      case 13:
        _align = <int>[6, 34, 62];
        break;
      case 14:
        _align = <int>[6, 26, 46, 66];
        break;
      case 15:
        _align = <int>[6, 26, 48, 70];
        break;
      case 16:
        _align = <int>[6, 26, 50, 74];
        break;
      case 17:
        _align = <int>[6, 30, 54, 78];
        break;
      case 18:
        _align = <int>[6, 30, 56, 82];
        break;
      case 19:
        _align = <int>[6, 30, 58, 86];
        break;
      case 20:
        _align = <int>[6, 34, 62, 90];
        break;
      case 21:
        _align = <int>[6, 28, 50, 72, 94];
        break;
      case 22:
        _align = <int>[6, 26, 50, 74, 98];
        break;
      case 23:
        _align = <int>[6, 30, 54, 78, 102];
        break;
      case 24:
        _align = <int>[6, 28, 54, 80, 106];
        break;
      case 25:
        _align = <int>[6, 32, 58, 84, 110];
        break;
      case 26:
        _align = <int>[6, 30, 58, 86, 114];
        break;
      case 27:
        _align = <int>[6, 34, 62, 90, 118];
        break;
      case 28:
        _align = <int>[6, 26, 50, 74, 98, 122];
        break;
      case 29:
        _align = <int>[6, 30, 54, 78, 102, 126];
        break;
      case 30:
        _align = <int>[6, 26, 52, 78, 104, 130];
        break;
      case 31:
        _align = <int>[6, 30, 56, 82, 108, 134];
        break;
      case 32:
        _align = <int>[6, 34, 60, 86, 112, 138];
        break;
      case 33:
        _align = <int>[6, 30, 58, 86, 114, 142];
        break;
      case 34:
        _align = <int>[6, 34, 62, 90, 118, 146];
        break;
      case 35:
        _align = <int>[6, 30, 54, 78, 102, 126, 150];
        break;
      case 36:
        _align = <int>[6, 24, 50, 76, 102, 128, 154];
        break;
      case 37:
        _align = <int>[6, 28, 54, 80, 106, 132, 158];
        break;
      case 38:
        _align = <int>[6, 32, 58, 84, 110, 136, 162];
        break;
      case 39:
        _align = <int>[6, 26, 54, 82, 110, 138, 166];
        break;
      case 40:
        _align = <int>[6, 30, 58, 86, 114, 142, 170];
        break;
    }
    return _align;
  }

  /// Converts the string to the list of bool
  List<bool> _stringToBoolArray(String _numberInString, int _noOfBits) {
    final List<bool> _numberInBool = List<bool>(_noOfBits);
    int j = 0;
    for (int i = 0; i < _numberInString.length; i++) {
      j = j * 10 + _numberInString[i].codeUnitAt(0) - 48;
    }

    for (int i = 0; i < _noOfBits; i++) {
      _numberInBool[_noOfBits - i - 1] = ((j >> i) & 1) == 1;
    }
    return _numberInBool;
  }

  /// Converts the integer to bool array value
  List<bool> _intToBoolArray(int _number, int _noOfBits) {
    final List<bool> _numberInBool = List<bool>(_noOfBits);
    for (int i = 0; i < _noOfBits; i++) {
      _numberInBool[_noOfBits - i - 1] = ((_number >> i) & 1) == 1;
    }

    return _numberInBool;
  }

  ///Methods to creates the block value based on the encoded data
  List<List<String>> _createBlocks(List<bool> _encodeData, int _noOfBlocks) {
    final List<List<String>> _encodedBlocks = List<List<String>>.generate(
        _noOfBlocks,
        (int i) => List<String>(_encodeData.length ~/ 8 ~/ _noOfBlocks));

    String _stringValue = '';
    int j = 0;
    int i = 0;
    int _blockNumber = 0;
    for (; j < _encodeData.length; j++) {
      if (j % 8 == 0 && j != 0) {
        _encodedBlocks[_blockNumber][i] = _stringValue;
        _stringValue = '';
        i++;

        if (i == (_encodeData.length / _noOfBlocks / 8)) {
          _blockNumber++;
          i = 0;
        }
      }

      _stringValue += _encodeData[j] ? '1' : '0';
    }

    _encodedBlocks[_blockNumber][i] = _stringValue;
    return _encodedBlocks;
  }

  /// Method to the split code word
  List<String> _splitCodeWord(
      List<List<String>> _encodeData, int _block, int _count) {
    final List<String> _encodeDataString = List<String>(_count);
    for (int i = 0; i < _count; i++) {
      _encodeDataString[i] = _encodeData[_block][i];
    }

    return _encodeDataString;
  }

  /// Method to find the actual mode, correction level and the input level
  void _initialize() {
    QRInputMode _actualMode = QRInputMode.numeric;
    for (int i = 0; i < _encodedText.length; i++) {
      if (_encodedText[i].codeUnitAt(0) < 58 &&
          _encodedText[i].codeUnitAt(0) > 47) {
      } else if ((_encodedText[i].codeUnitAt(0) < 91 &&
              _encodedText[i].codeUnitAt(0) > 64) ||
          _encodedText[i] == '\$' ||
          _encodedText[i] == '%' ||
          _encodedText[i] == '*' ||
          _encodedText[i] == '+' ||
          _encodedText[i] == '-' ||
          _encodedText[i] == '.' ||
          _encodedText[i] == '/' ||
          _encodedText[i] == ':' ||
          _encodedText[i] == ' ') {
        _actualMode = QRInputMode.alphaNumeric;
      } else if ((_encodedText[i].codeUnitAt(0) >= 65377 &&
              _encodedText[i].codeUnitAt(0) <= 65439) ||
          (_encodedText[i].codeUnitAt(0) >= 97 &&
              _encodedText[i].codeUnitAt(0) <= 122)) {
        _actualMode = QRInputMode.binary;

        break;
      } else {
        _actualMode = QRInputMode.binary;
        _isEci = true;
        break;
      }
    }

    if (_isUserMentionedMode) {
      if (_actualMode != inputMode) {
        if (((_actualMode == QRInputMode.alphaNumeric ||
                    _actualMode == QRInputMode.binary) &&
                inputMode == QRInputMode.numeric) ||
            (_actualMode == QRInputMode.binary &&
                inputMode == QRInputMode.alphaNumeric)) {
          // new BarcodeException("Mode Conflict");
        }
      }
    }

    _inputMode = _actualMode;
    _initializeECI();
    _initializeVersionAndECR();
    _noOfModules = (_getVersionNumber(_codeVersion) - 1) * 4 + 21;
  }

  /// Method to initialize the error correction level
  void _initializeECI() {
    if (_isEci) {
      for (int i = 0; i < _encodedText.length; i++) {
        if (_encodedText[i].codeUnitAt(0) >= 32 &&
            _encodedText[i].codeUnitAt(0) <= 255) {
          continue;
        }
        //Check for CP437
        if (_isCP437Character(_encodedText[i].codeUnitAt(0))) {
          _eciAssignmentNumber = 2;
          break;
        }

        //Check for ISO/IEC 8859-2
        else if (_isISO8859_2CharacterSet(_encodedText[i].codeUnitAt(0))) {
          _eciAssignmentNumber = 4;
          break;
        }
        //Check for ISO/IEC 8859-3
        else if (_isISO8859_3CharacterSet(_encodedText[i].codeUnitAt(0))) {
          _eciAssignmentNumber = 5;
          break;
        }
        //Check for ISO/IEC 8859-4
        else if (_isISO8859_4CharacterSet(_encodedText[i].codeUnitAt(0))) {
          _eciAssignmentNumber = 6;
          break;
        }
        //Check for ISO/IEC 8859-5
        else if (_isISO8859_5CharacterSet(_encodedText[i].codeUnitAt(0))) {
          _eciAssignmentNumber = 7;
          break;
        }
        //Check for ISO/IEC 8859-6
        else if (_isISO8859_6CharacterSet(_encodedText[i].codeUnitAt(0))) {
          _eciAssignmentNumber = 8;
          break;
        }
        //Check for ISO/IEC 8859-7
        else if (_isISO8859_7CharacterSet(_encodedText[i].codeUnitAt(0))) {
          _eciAssignmentNumber = 9;
          break;
        }
        //Check for ISO/IEC 8859-8
        else if (_isISO8859_8CharacterSet(_encodedText[i].codeUnitAt(0))) {
          _eciAssignmentNumber = 10;
          break;
        }
        //Check for ISO/IEC 8859-8
        else if (_isISO8859_11CharacterSet(_encodedText[i].codeUnitAt(0))) {
          _eciAssignmentNumber = 13;
          break;
        }

        //Check for Windows1250
        else if (_isWindows1250Character(_encodedText[i].codeUnitAt(0))) {
          _eciAssignmentNumber = 21;
          break;
        }
        //Check for Windows1251
        else if (_isWindows1251Character(_encodedText[i].codeUnitAt(0))) {
          _eciAssignmentNumber = 22;
          break;
        }
        //Check for Windows1252
        else if (_isWindows1252Character(_encodedText[i].codeUnitAt(0))) {
          _eciAssignmentNumber = 23;
          break;
        }
        //Check for Windows1256
        else if (_isWindows1256Character(_encodedText[i].codeUnitAt(0))) {
          _eciAssignmentNumber = 24;
          break;
        }
      }
    }
  }

  /// Method to initialize the version and the error correction level
  void _initializeVersionAndECR() {
    if (!_isUserMentionedVersion || _codeVersion == QRCodeVersion.auto) {
      List<int> _dataCapacityOfVersions;
      if (_isUserMentionedErrorCorrectionLevel) {
        switch (_inputMode) {
          case QRInputMode.numeric:
            {
              switch (_errorCorrectionLevel) {
                case ErrorCorrectionLevel.low:
                  _dataCapacityOfVersions =
                      _QRCodeValue._numericDataCapacityLow;
                  break;
                case ErrorCorrectionLevel.medium:
                  _dataCapacityOfVersions =
                      _QRCodeValue._numericDataCapacityMedium;
                  break;
                case ErrorCorrectionLevel.quartile:
                  _dataCapacityOfVersions =
                      _QRCodeValue._numericDataCapacityQuartile;
                  break;
                case ErrorCorrectionLevel.high:
                  _dataCapacityOfVersions =
                      _QRCodeValue._numericDataCapacityHigh;
                  break;
              }
              break;
            }

          case QRInputMode.alphaNumeric:
            {
              switch (_errorCorrectionLevel) {
                case ErrorCorrectionLevel.low:
                  _dataCapacityOfVersions =
                      _QRCodeValue._alphaNumericDataCapacityLow;
                  break;
                case ErrorCorrectionLevel.medium:
                  _dataCapacityOfVersions =
                      _QRCodeValue._alphaNumericDataCapacityMedium;
                  break;
                case ErrorCorrectionLevel.quartile:
                  _dataCapacityOfVersions =
                      _QRCodeValue._alphaNumericDataCapacityQuartile;
                  break;
                case ErrorCorrectionLevel.high:
                  _dataCapacityOfVersions =
                      _QRCodeValue._alphaNumericDataCapacityHigh;
                  break;
              }
              break;
            }

          case QRInputMode.binary:
            {
              switch (_errorCorrectionLevel) {
                case ErrorCorrectionLevel.low:
                  _dataCapacityOfVersions = _QRCodeValue._binaryDataCapacityLow;
                  break;
                case ErrorCorrectionLevel.medium:
                  _dataCapacityOfVersions =
                      _QRCodeValue._binaryDataCapacityMedium;
                  break;
                case ErrorCorrectionLevel.quartile:
                  _dataCapacityOfVersions =
                      _QRCodeValue._binaryDataCapacityQuartile;
                  break;
                case ErrorCorrectionLevel.high:
                  _dataCapacityOfVersions =
                      _QRCodeValue._binaryDataCapacityHigh;
                  break;
              }
              break;
            }
        }
      } else {
        _errorCorrectionLevel = ErrorCorrectionLevel.low;
        switch (_inputMode) {
          case QRInputMode.numeric:
            {
              _dataCapacityOfVersions = _QRCodeValue._numericDataCapacityLow;
              break;
            }

          case QRInputMode.alphaNumeric:
            {
              _dataCapacityOfVersions =
                  _QRCodeValue._alphaNumericDataCapacityLow;
              break;
            }

          case QRInputMode.binary:
            {
              _dataCapacityOfVersions = _QRCodeValue._binaryDataCapacityLow;
              break;
            }
        }
      }
      int i;
      for (i = 0; i < _dataCapacityOfVersions.length; i++) {
        if (_dataCapacityOfVersions[i] > _encodedText.length) {
          break;
        }
      }

      final int _calculatedVersion = i + 1;

      if (_calculatedVersion > 40) {
        return;
      } else {
        _codeVersion = _getVersionBasedOnNumber(_calculatedVersion);
      }
    } else if (_isUserMentionedVersion) {
      if (_isUserMentionedErrorCorrectionLevel) {
        int capacity = 0;
        if (_inputMode == QRInputMode.alphaNumeric) {
          capacity = _QRCodeValue._getAlphaNumericDataCapacity(
              _errorCorrectionLevel, _codeVersion);
        } else if (_inputMode == QRInputMode.numeric) {
          capacity = _QRCodeValue._getNumericDataCapacity(
              _errorCorrectionLevel, _codeVersion);
        } else if (_inputMode == QRInputMode.binary) {
          capacity = _QRCodeValue._getBinaryDataCapacity(
              _errorCorrectionLevel, _codeVersion);
        }
        if (capacity < _encodedText.length) {
          throw 'The input value length is greater than version capacity';
        }
      } else {
        int capacityLow = 0,
            capacityMedium = 0,
            capacityQuartile = 0,
            capacityHigh = 0;
        if (_inputMode == QRInputMode.alphaNumeric) {
          capacityLow = _QRCodeValue._getAlphaNumericDataCapacity(
              ErrorCorrectionLevel.low, _codeVersion);
          capacityMedium = _QRCodeValue._getAlphaNumericDataCapacity(
              ErrorCorrectionLevel.medium, _codeVersion);
          capacityQuartile = _QRCodeValue._getAlphaNumericDataCapacity(
              ErrorCorrectionLevel.quartile, _codeVersion);
          capacityHigh = _QRCodeValue._getAlphaNumericDataCapacity(
              ErrorCorrectionLevel.high, _codeVersion);
        } else if (_inputMode == QRInputMode.numeric) {
          capacityLow = _QRCodeValue._getNumericDataCapacity(
              ErrorCorrectionLevel.low, _codeVersion);
          capacityMedium = _QRCodeValue._getNumericDataCapacity(
              ErrorCorrectionLevel.medium, _codeVersion);
          capacityQuartile = _QRCodeValue._getNumericDataCapacity(
              ErrorCorrectionLevel.quartile, _codeVersion);
          capacityHigh = _QRCodeValue._getNumericDataCapacity(
              ErrorCorrectionLevel.high, _codeVersion);
        } else if (_inputMode == QRInputMode.binary) {
          capacityLow = _QRCodeValue._getBinaryDataCapacity(
              ErrorCorrectionLevel.low, _codeVersion);
          capacityMedium = _QRCodeValue._getBinaryDataCapacity(
              ErrorCorrectionLevel.medium, _codeVersion);
          capacityQuartile = _QRCodeValue._getBinaryDataCapacity(
              ErrorCorrectionLevel.quartile, _codeVersion);
          capacityHigh = _QRCodeValue._getBinaryDataCapacity(
              ErrorCorrectionLevel.high, _codeVersion);
        }

        if (capacityHigh > _encodedText.length) {
          _errorCorrectionLevel = ErrorCorrectionLevel.high;
        } else if (capacityQuartile > _encodedText.length) {
          _errorCorrectionLevel = ErrorCorrectionLevel.quartile;
        } else if (capacityMedium > _encodedText.length) {
          _errorCorrectionLevel = ErrorCorrectionLevel.medium;
        } else if (capacityLow > _encodedText.length) {
          _errorCorrectionLevel = ErrorCorrectionLevel.low;
        } else {
          throw 'The input value length is greater than version capacity';
        }
      }
    }
  }

  /// Method to draw the format information
  void _drawFormatInformation() {
    final List<int> _formatInformation = _qrCodeValues._formatInformation;
    int _count = 0;
    for (int i = 0; i < 7; i++) {
      if (i == 6) {
        _moduleValues[i + 1][8]._isBlack =
            _formatInformation[_count] == 1 ? true : false;
      } else {
        _moduleValues[i][8]._isBlack =
            _formatInformation[_count] == 1 ? true : false;
      }

      _moduleValues[8][_noOfModules - i - 1]._isBlack =
          _formatInformation[_count++] == 1 ? true : false;
    }

    _count = 14;

    for (int i = 0; i < 7; i++) {
      //Draw format information from 0 to 6
      if (i == 6) {
        _moduleValues[8][i + 1]._isBlack =
            _formatInformation[_count] == 1 ? true : false;
      } else {
        _moduleValues[8][i]._isBlack =
            _formatInformation[_count] == 1 ? true : false;
      }

      _moduleValues[_noOfModules - i - 1][8]._isBlack =
          _formatInformation[_count--] == 1 ? true : false;
    }

    //Draw format information 7
    _moduleValues[8][8]._isBlack = _formatInformation[7] == 1 ? true : false;
    _moduleValues[8][_noOfModules - 8]._isBlack =
        _formatInformation[7] == 1 ? true : false;
  }

  /// Method used for data allocation and masking
  void _dataAllocationAndMasking(List<bool> _data) {
    _dataAllocationValues = List<List<_ModuleValue>>.generate(
        _noOfModules,
        (int i) => List<_ModuleValue>.generate(
            _noOfModules, (int j) => _ModuleValue()));

    int _point = 0;

    for (int i = _noOfModules - 1; i >= 0; i -= 2) {
      for (int j = _noOfModules - 1; j >= 0; j--) {
        if (!(_moduleValues[i][j]._isFilled &&
            _moduleValues[i - 1][j]._isFilled)) {
          if (!_moduleValues[i][j]._isFilled) {
            if (_point + 1 < _data.length) {
              _dataAllocationValues[i][j]._isBlack = _data[_point++];
            }

            if ((i + j) % 3 == 0) {
              if (_dataAllocationValues[i][j]._isBlack)
                _dataAllocationValues[i][j]._isBlack = true;
              else
                _dataAllocationValues[i][j]._isBlack = false;
            } else {
              if (_dataAllocationValues[i][j]._isBlack)
                _dataAllocationValues[i][j]._isBlack = false;
              else
                _dataAllocationValues[i][j]._isBlack = true;
            }

            _dataAllocationValues[i][j]._isFilled = true;
          }

          if (!_moduleValues[i - 1][j]._isFilled) {
            if (_point + 1 < _data.length) {
              _dataAllocationValues[i - 1][j]._isBlack = _data[_point++];
            }

            if ((i - 1 + j) % 3 == 0) {
              if (_dataAllocationValues[i - 1][j]._isBlack)
                _dataAllocationValues[i - 1][j]._isBlack = true;
              else
                _dataAllocationValues[i - 1][j]._isBlack = false;
            } else {
              if (_dataAllocationValues[i - 1][j]._isBlack)
                _dataAllocationValues[i - 1][j]._isBlack = false;
              else
                _dataAllocationValues[i - 1][j]._isBlack = true;
            }
            _dataAllocationValues[i - 1][j]._isFilled = true;
          }
        }
      }

      i -= 2;
      if (i == 6) {
        i--;
      }

      for (int j = 0; j < _noOfModules; j++) {
        if (!(_moduleValues[i][j]._isFilled &&
            _moduleValues[i - 1][j]._isFilled)) {
          if (!_moduleValues[i][j]._isFilled) {
            if (_point + 1 < _data.length)
              _dataAllocationValues[i][j]._isBlack = _data[_point++];
            if ((i + j) % 3 == 0) {
              if (_dataAllocationValues[i][j]._isBlack)
                _dataAllocationValues[i][j]._isBlack = true;
              else
                _dataAllocationValues[i][j]._isBlack = false;
            } else {
              if (_dataAllocationValues[i][j]._isBlack)
                _dataAllocationValues[i][j]._isBlack = false;
              else
                _dataAllocationValues[i][j]._isBlack = true;
            }
            _dataAllocationValues[i][j]._isFilled = true;
          }

          if (!_moduleValues[i - 1][j]._isFilled) {
            if (_point + 1 < _data.length)
              _dataAllocationValues[i - 1][j]._isBlack = _data[_point++];
            if ((i - 1 + j) % 3 == 0) {
              if (_dataAllocationValues[i - 1][j]._isBlack)
                _dataAllocationValues[i - 1][j]._isBlack = true;
              else
                _dataAllocationValues[i - 1][j]._isBlack = false;
            } else {
              if (_dataAllocationValues[i - 1][j]._isBlack)
                _dataAllocationValues[i - 1][j]._isBlack = false;
              else
                _dataAllocationValues[i - 1][j]._isBlack = true;
            }
            _dataAllocationValues[i - 1][j]._isFilled = true;
          }
        }
      }
    }
    for (int i = 0; i < _noOfModules; i++) {
      for (int j = 0; j < _noOfModules; j++) {
        if (!_moduleValues[i][j]._isFilled) {
          final bool _flag = _dataAllocationValues[i][j]._isBlack;
          if (_flag)
            _dataAllocationValues[i][j]._isBlack = false;
          else
            _dataAllocationValues[i][j]._isBlack = true;
        }
      }
    }
  }

  /// Method to draw the alignment pattern
  void _drawAlignmentPattern(int x, int y) {
    int i = x - 2, j = y - 2;
    for (; i < x + 3; i++, j++) {
      _moduleValues[i][y - 2]._isBlack = true;
      _moduleValues[i][y - 2]._isFilled = true;

      _moduleValues[i][y + 2]._isBlack = true;
      _moduleValues[i][y + 2]._isFilled = true;

      _moduleValues[x - 2][j]._isBlack = true;
      _moduleValues[x - 2][j]._isFilled = true;

      _moduleValues[x + 2][j]._isBlack = true;
      _moduleValues[x + 2][j]._isFilled = true;
    }

    i = x - 1;
    j = y - 1;
    for (; i < x + 2; i++, j++) {
      _moduleValues[i][y - 1]._isBlack = false;
      _moduleValues[i][y - 1]._isFilled = true;

      _moduleValues[i][y + 1]._isBlack = false;
      _moduleValues[i][y + 1]._isFilled = true;

      _moduleValues[x - 1][j]._isBlack = false;
      _moduleValues[x - 1][j]._isFilled = true;

      _moduleValues[x + 1][j]._isBlack = false;
      _moduleValues[x + 1][j]._isFilled = true;
    }

    _moduleValues[x][y]._isBlack = true;
    _moduleValues[x][y]._isFilled = true;
  }

  /// Method to draw the timing pattern
  void _drawTimingPattern() {
    for (int i = 8; i < _noOfModules - 8; i += 2) {
      _moduleValues[i][6]._isBlack = true;
      _moduleValues[i][6]._isFilled = true;

      _moduleValues[i + 1][6]._isBlack = false;
      _moduleValues[i + 1][6]._isFilled = true;

      _moduleValues[6][i]._isBlack = true;
      _moduleValues[6][i]._isFilled = true;

      _moduleValues[6][i + 1]._isBlack = false;
      _moduleValues[6][i + 1]._isFilled = true;
    }

    _moduleValues[_noOfModules - 8][8]._isBlack = true;
    _moduleValues[_noOfModules - 8][8]._isFilled = true;
  }

  /// Method to draw the PDP
  void _drawPDP(int x, int y) {
    int i = x;
    int j = y;
    for (; i < x + 7; i++, j++) {
      _moduleValues[i][y]._isBlack = true;
      _moduleValues[i][y]._isFilled = true;
      _moduleValues[i][y]._isPDP = true;

      _moduleValues[i][y + 6]._isBlack = true;
      _moduleValues[i][y + 6]._isFilled = true;
      _moduleValues[i][y + 6]._isPDP = true;

      if (y + 7 < _noOfModules) {
        _moduleValues[i][y + 7]._isBlack = false;
        _moduleValues[i][y + 7]._isFilled = true;
        _moduleValues[i][y + 7]._isPDP = true;
      } else if (y - 1 >= 0) {
        _moduleValues[i][y - 1]._isBlack = false;
        _moduleValues[i][y - 1]._isFilled = true;
        _moduleValues[i][y - 1]._isPDP = true;
      }

      _moduleValues[x][j]._isBlack = true;
      _moduleValues[x][j]._isFilled = true;
      _moduleValues[x][j]._isPDP = true;

      _moduleValues[x + 6][j]._isBlack = true;
      _moduleValues[x + 6][j]._isFilled = true;
      _moduleValues[x + 6][j]._isPDP = true;

      if (x + 7 < _noOfModules) {
        _moduleValues[x + 7][j]._isBlack = false;
        _moduleValues[x + 7][j]._isFilled = true;
        _moduleValues[x + 7][j]._isPDP = true;
      } else if (x - 1 >= 0) {
        _moduleValues[x - 1][j]._isBlack = false;
        _moduleValues[x - 1][j]._isFilled = true;
        _moduleValues[x - 1][j]._isPDP = true;
      }
    }

    if (x + 7 < _noOfModules && y + 7 < _noOfModules) {
      _moduleValues[x + 7][y + 7]._isBlack = false;
      _moduleValues[x + 7][y + 7]._isFilled = true;
      _moduleValues[x + 7][y + 7]._isPDP = true;
    } else if (x + 7 < _noOfModules && y + 7 >= _noOfModules) {
      _moduleValues[x + 7][y - 1]._isBlack = false;
      _moduleValues[x + 7][y - 1]._isFilled = true;
      _moduleValues[x + 7][y - 1]._isPDP = true;
    } else if (x + 7 >= _noOfModules && y + 7 < _noOfModules) {
      _moduleValues[x - 1][y + 7]._isBlack = false;
      _moduleValues[x - 1][y + 7]._isFilled = true;
      _moduleValues[x - 1][y + 7]._isPDP = true;
    }

    x++;
    y++;
    i = x;
    j = y;
    for (; i < x + 5; i++, j++) {
      _moduleValues[i][y]._isBlack = false;
      _moduleValues[i][y]._isFilled = true;
      _moduleValues[i][y]._isPDP = true;

      _moduleValues[i][y + 4]._isBlack = false;
      _moduleValues[i][y + 4]._isFilled = true;
      _moduleValues[i][y + 4]._isPDP = true;

      _moduleValues[x][j]._isBlack = false;
      _moduleValues[x][j]._isFilled = true;
      _moduleValues[x][j]._isPDP = true;

      _moduleValues[x + 4][j]._isBlack = false;
      _moduleValues[x + 4][j]._isFilled = true;
      _moduleValues[x + 4][j]._isPDP = true;
    }

    x++;
    y++;
    i = x;
    j = y;
    for (; i < x + 3; i++, j++) {
      _moduleValues[i][y]._isBlack = true;
      _moduleValues[i][y]._isFilled = true;
      _moduleValues[i][y]._isPDP = true;

      _moduleValues[i][y + 2]._isBlack = true;
      _moduleValues[i][y + 2]._isFilled = true;
      _moduleValues[i][y + 2]._isPDP = true;

      _moduleValues[x][j]._isBlack = true;
      _moduleValues[x][j]._isFilled = true;
      _moduleValues[x][j]._isPDP = true;

      _moduleValues[x + 2][j]._isBlack = true;
      _moduleValues[x + 2][j]._isFilled = true;
      _moduleValues[x + 2][j]._isPDP = true;
    }

    _moduleValues[x + 1][y + 1]._isBlack = true;
    _moduleValues[x + 1][y + 1]._isFilled = true;
    _moduleValues[x + 1][y + 1]._isPDP = true;
  }

  /// Method used to generate the value
  void _generateValues() {
    _initialize();
    _qrCodeValues = _QRCodeValue(
        qrCodeVersion: _codeVersion,
        errorCorrectionLevel: _errorCorrectionLevel);
    _moduleValues = List<List<_ModuleValue>>.generate(
        _noOfModules,
        (int i) => List<_ModuleValue>.generate(
            _noOfModules, (int j) => _ModuleValue()));
    _drawPDP(0, 0);
    _drawPDP(_noOfModules - 7, 0);
    _drawPDP(0, _noOfModules - 7);
    _drawTimingPattern();
    if (_codeVersion != QRCodeVersion.version01) {
      final List<int> _allignCoOrdinates = _getAlignmentPatternCoordinates();

      for (int i = 0; i < _allignCoOrdinates.length; i++) {
        for (int j = 0; j < _allignCoOrdinates.length; j++) {
          if (!_moduleValues[_allignCoOrdinates[i]][_allignCoOrdinates[j]]
              ._isPDP) {
            _drawAlignmentPattern(_allignCoOrdinates[i], _allignCoOrdinates[j]);
          }
        }
      }
    }

    _allocateFormatAndVersionInformation();
    _encodeDataCodeWords = _encodeData();
    _dataAllocationAndMasking(_encodeDataCodeWords);
    _drawFormatInformation();
  }

  /// Method to encoded the data based on mode
  void _encodeDataBasedOnMode() {
    switch (_inputMode) {
      case QRInputMode.numeric:
        _encodeDataCodeWords.add(false);
        _encodeDataCodeWords.add(false);
        _encodeDataCodeWords.add(false);
        _encodeDataCodeWords.add(true);
        break;

      case QRInputMode.alphaNumeric:
        _encodeDataCodeWords.add(false);
        _encodeDataCodeWords.add(false);
        _encodeDataCodeWords.add(true);
        _encodeDataCodeWords.add(false);
        break;

      case QRInputMode.binary:
        if (_isEci) {
          //Add ECI Mode Indicator
          _encodeDataCodeWords.add(false);
          _encodeDataCodeWords.add(true);
          _encodeDataCodeWords.add(true);
          _encodeDataCodeWords.add(true);

          //Add ECI assignment number
          final List<bool> numberInBool =
              _stringToBoolArray(_eciAssignmentNumber.toString(), 8);
          for (int i = 0; i < numberInBool.length; i++) {
            _encodeDataCodeWords.add(numberInBool[i]);
          }
        }
        _encodeDataCodeWords.add(false);
        _encodeDataCodeWords.add(true);
        _encodeDataCodeWords.add(false);
        _encodeDataCodeWords.add(false);
        break;
    }
  }

  /// Method to set the indicator count for lower version
  int _setIndicatorCountForLowerVersion() {
    int numberOfBitsInCharacterCountIndicator = 0;
    switch (_inputMode) {
      case QRInputMode.numeric:
        numberOfBitsInCharacterCountIndicator = 10;
        break;

      case QRInputMode.alphaNumeric:
        numberOfBitsInCharacterCountIndicator = 9;
        break;
      case QRInputMode.binary:
        numberOfBitsInCharacterCountIndicator = 8;
        break;
    }

    return numberOfBitsInCharacterCountIndicator;
  }

  /// method to set the indicator count for higher version
  int _setIndicatorCountForHigherVersion() {
    int numberOfBitsInCharacterCountIndicator = 0;
    switch (_inputMode) {
      case QRInputMode.numeric:
        numberOfBitsInCharacterCountIndicator = 12;
        break;

      case QRInputMode.alphaNumeric:
        numberOfBitsInCharacterCountIndicator = 11;
        break;

      case QRInputMode.binary:
        numberOfBitsInCharacterCountIndicator = 16;
        break;
    }

    return numberOfBitsInCharacterCountIndicator;
  }

  /// Method to set the indicator count
  int _setIndicatorCount() {
    int numberOfBitsInCharacterCountIndicator = 0;
    switch (_inputMode) {
      case QRInputMode.numeric:
        numberOfBitsInCharacterCountIndicator = 14;
        break;

      case QRInputMode.alphaNumeric:
        numberOfBitsInCharacterCountIndicator = 13;
        break;

      case QRInputMode.binary:
        numberOfBitsInCharacterCountIndicator = 16;
        break;
    }
    return numberOfBitsInCharacterCountIndicator;
  }

  /// Method to encode the data of numeric mode
  void _encodeDataInNumericMode() {
    String number = '';
    for (int i = 0; i < _encodedText.length; i++) {
      List<bool> numberInBool;
      number += _encodedText[i];

      if (i % 3 == 2 && i != 0 || i == _encodedText.length - 1) {
        if (number.length == 3)
          numberInBool = _stringToBoolArray(number, 10);
        else if (number.length == 2)
          numberInBool = _stringToBoolArray(number, 7);
        else
          numberInBool = _stringToBoolArray(number, 4);
        number = '';
        for (int i = 0; i < numberInBool.length; i++) {
          _encodeDataCodeWords.add(numberInBool[i]);
        }
      }
    }
  }

  /// Method to encode the alpha numeric data
  void _encodeDataForAlphaNumeric() {
    String numberInString = '';
    int number = 0;
    for (int i = 0; i < _encodedText.length; i++) {
      List<bool> numberInBool;
      numberInString += _encodedText[i];

      if (i % 2 == 0 && i + 1 != _encodedText.length)
        number = 45 * _qrCodeValues._getAlphaNumericValues(_encodedText[i]);

      if (i % 2 == 1 && i != 0) {
        number += _qrCodeValues._getAlphaNumericValues(_encodedText[i]);
        numberInBool = _intToBoolArray(number, 11);
        number = 0;
        for (int i = 0; i < numberInBool.length; i++) {
          _encodeDataCodeWords.add(numberInBool[i]);
        }

        numberInString = '';
      }
      if (i != 1 && numberInString != null) {
        if (i + 1 == _encodedText.length && numberInString.length == 1) {
          number = _qrCodeValues._getAlphaNumericValues(_encodedText[i]);
          numberInBool = _intToBoolArray(number, 6);
          number = 0;
          for (int i = 0; i < numberInBool.length; i++) {
            _encodeDataCodeWords.add(numberInBool[i]);
          }
        }
      }
    }
  }

  /// method to encode the binary data
  void _encodeDataForBinary() {
    for (int i = 0; i < _encodedText.length; i++) {
      int number = 0;
      if ((_encodedText[i].codeUnitAt(0) >= 32 &&
              _encodedText[i].codeUnitAt(0) <= 126) ||
          _encodedText[i].codeUnitAt(0) >= 161 &&
              _encodedText[i].codeUnitAt(0) <= 255 ||
          _encodedText[i].codeUnitAt(0) == 10 ||
          _encodedText[i].codeUnitAt(0) == 13) {
        number = _encodedText[i].codeUnitAt(0);
      } else if (_encodedText[i].codeUnitAt(0) >= 65377 &&
          _encodedText[i].codeUnitAt(0) <= 65439) {
        number = _encodedText[i].codeUnitAt(0) - 65216;
      } else if (_encodedText[i].codeUnitAt(0) >= 1025 &&
          _encodedText[i].codeUnitAt(0) <= 1119) {
        number = _encodedText[i].codeUnitAt(0) - 864;
      } else if (_encodedText[i].codeUnitAt(0) >= 260 &&
          _encodedText[i].codeUnitAt(0) <= 382) {
        //   Encoding europeanEncode= Encoding.GetEncoding("ISO-8859-2");
//      byte[] charCode= europeanEncode.GetBytes(Text[i].ToString());
//      number = charCode[0];
      } else {
        throw 'The provided input value contains non-convertible characters';
      }

      final List<bool> numberInBool = _intToBoolArray(number, 8);
      for (int i = 0; i < numberInBool.length; i++) {
        _encodeDataCodeWords.add(numberInBool[i]);
      }
    }
  }

  /// Method to encode the code words
  void _encodeCodeWords() {
    for (int i = 0; i < 4; i++) {
      if (_encodeDataCodeWords.length / 8 == _qrCodeValues._noOfDataCodeWord)
        break;
      else
        _encodeDataCodeWords.add(false);
    }

    for (;;) {
      if (_encodeDataCodeWords.length % 8 == 0)
        break;
      else
        _encodeDataCodeWords.add(false);
    }

    for (;;) {
      if (_encodeDataCodeWords.length / 8 == _qrCodeValues._noOfDataCodeWord)
        break;
      else {
        _encodeDataCodeWords.add(true);
        _encodeDataCodeWords.add(true);
        _encodeDataCodeWords.add(true);
        _encodeDataCodeWords.add(false);
        _encodeDataCodeWords.add(true);
        _encodeDataCodeWords.add(true);
        _encodeDataCodeWords.add(false);
        _encodeDataCodeWords.add(false);
      }
      if (_encodeDataCodeWords.length / 8 == _qrCodeValues._noOfDataCodeWord)
        break;
      else {
        _encodeDataCodeWords.add(false);
        _encodeDataCodeWords.add(false);
        _encodeDataCodeWords.add(false);
        _encodeDataCodeWords.add(true);
        _encodeDataCodeWords.add(false);
        _encodeDataCodeWords.add(false);
        _encodeDataCodeWords.add(false);
        _encodeDataCodeWords.add(true);
      }
    }

    _dataBits = _qrCodeValues._noOfDataCodeWord;
    _blocks = _qrCodeValues._noOfErrorCorrectionBlocks;

    int totalBlockSize = _blocks[0];
    if (_blocks.length == 6) {
      totalBlockSize = _blocks[0] + _blocks[3];
    }
    final List<List<String>> _ds1 =
        List<List<String>>.generate(totalBlockSize, (int i) => <String>[]);

    List<bool> testEncodeData = _encodeDataCodeWords;
    if (_blocks.length == 6) {
      final int dataCodeWordLength = _blocks[0] * _blocks[2] * 8;
      testEncodeData = <bool>[];
      for (int i = 0; i < dataCodeWordLength; i++) {
        testEncodeData.add(_encodeDataCodeWords[i]);
      }
    }

    List<List<String>> _dsOne = List<List<String>>.generate(_blocks[0],
        (int i) => List<String>(testEncodeData.length ~/ 8 ~/ _blocks[0]));
    _dsOne = _createBlocks(testEncodeData, _blocks[0]);

    for (int i = 0; i < _blocks[0]; i++) {
      _ds1[i] =
          _splitCodeWord(_dsOne, i, testEncodeData.length ~/ 8 ~/ _blocks[0]);
    }

    if (_blocks.length == 6) {
      testEncodeData = <bool>[];
      for (int i = _blocks[0] * _blocks[2] * 8;
          i < _encodeDataCodeWords.length;
          i++) {
        testEncodeData.add(_encodeDataCodeWords[i]);
      }

      List<List<String>> _dsTwo = List<List<String>>.generate(_blocks[0],
          (int i) => List<String>(testEncodeData.length ~/ 8 ~/ _blocks[3]));
      _dsTwo = _createBlocks(testEncodeData, _blocks[3]);

      for (int i = _blocks[0], count = 0; i < totalBlockSize; i++) {
        _ds1[i] = _splitCodeWord(
            _dsTwo, count++, testEncodeData.length ~/ 8 ~/ _blocks[3]);
      }
    }

    _encodeDataCodeWords = <bool>[];
    for (int i = 0; i < 125; i++) {
      for (int k = 0; k < totalBlockSize; k++) {
        for (int j = 0; j < 8; j++) {
          if (i < _ds1[k].length)
            _encodeDataCodeWords.add(_ds1[k][i][j] == '1' ? true : false);
        }
      }
    }

    _calculateErrorCorrectingCodeWord(totalBlockSize, _ds1);
  }

  /// Method to encode the data
  List<bool> _encodeData() {
    _encodeDataCodeWords = <bool>[];
    _encodeDataBasedOnMode();

    int numberOfBitsInCharacterCountIndicator = 0;
    final int _currentVersion = _getVersionNumber(_codeVersion);
    if (_currentVersion < 10) {
      numberOfBitsInCharacterCountIndicator =
          _setIndicatorCountForLowerVersion();
    } else if (_currentVersion < 27) {
      numberOfBitsInCharacterCountIndicator =
          _setIndicatorCountForHigherVersion();
    } else {
      numberOfBitsInCharacterCountIndicator = _setIndicatorCount();
    }

    final List<bool> numberOfBitsInCharacterCountIndicatorInBool =
        _intToBoolArray(
            _encodedText.length, numberOfBitsInCharacterCountIndicator);

    for (int i = 0; i < numberOfBitsInCharacterCountIndicator; i++) {
      _encodeDataCodeWords.add(numberOfBitsInCharacterCountIndicatorInBool[i]);
    }

    if (_inputMode == QRInputMode.numeric) {
      _encodeDataInNumericMode();
    } else if (_inputMode == QRInputMode.alphaNumeric) {
      _encodeDataForAlphaNumeric();
    } else if (_inputMode == QRInputMode.binary) {
      _encodeDataForBinary();
    }

    _encodeCodeWords();
    return _encodeDataCodeWords;
  }

  /// Method to calculate the error correcting code word
  void _calculateErrorCorrectingCodeWord(
      int totalBlockSize, List<List<String>> _ds1) {
    final _ErrorCorrectionCodeWords _errorCorrectionCodeWord =
        _ErrorCorrectionCodeWords(
            codeVersion: _codeVersion, correctionLevel: _errorCorrectionLevel);

    _dataBits = _qrCodeValues._noOfDataCodeWord;
    final int _eccw = _qrCodeValues._noOfErrorCorrectionCodeWord;
    _blocks = _qrCodeValues._noOfErrorCorrectionBlocks;

    if (_blocks.length == 6) {
      _errorCorrectionCodeWord._dataBits =
          (_dataBits - _blocks[3] * _blocks[5]) ~/ _blocks[0];
    } else {
      _errorCorrectionCodeWord._dataBits = _dataBits ~/ _blocks[0];
    }

    _errorCorrectionCodeWord._eccw = _eccw ~/ totalBlockSize;

    final List<List<String>> _polynomial =
        List<List<String>>.generate(totalBlockSize, (int i) => <String>[]);
    int count = 0;

    for (int i = 0; i < _blocks[0]; i++) {
      _errorCorrectionCodeWord._dataCodeWords = _ds1[count];
      _polynomial[count++] = _errorCorrectionCodeWord._getERCW();
    }
    if (_blocks.length == 6) {
      _errorCorrectionCodeWord._dataBits =
          (_dataBits - _blocks[0] * _blocks[2]) ~/ _blocks[3];

      for (int i = 0; i < _blocks[3]; i++) {
        _errorCorrectionCodeWord._dataCodeWords = _ds1[count];
        _polynomial[count++] = _errorCorrectionCodeWord._getERCW();
      }
    }

    if (_blocks.length != 6) {
      for (int i = 0; i < _polynomial[0].length; i++) {
        for (int k = 0; k < _blocks[0]; k++)
          for (int j = 0; j < 8; j++)
            if (i < _polynomial[k].length) {
              _encodeDataCodeWords
                  .add(_polynomial[k][i][j] == '1' ? true : false);
            }
      }
    } else {
      for (int i = 0; i < _polynomial[0].length; i++) {
        for (int k = 0; k < totalBlockSize; k++)
          for (int j = 0; j < 8; j++)
            if (i < _polynomial[k].length)
              _encodeDataCodeWords
                  .add(_polynomial[k][i][j] == '1' ? true : false);
      }
    }
  }

  @override
  void _renderBarcode(
      Canvas canvas,
      Size size,
      Offset offset,
      String value,
      Color foregroundColor,
      TextStyle textStyle,
      double textSpacing,
      TextAlign textAlign,
      bool showValue) {
    _encodedText = value;
    _generateValues();

    int x = 0;
    final int w = _noOfModules;
    final int h = _noOfModules;

    final Paint _paint = _getBarPaint(foregroundColor);

    final double _minSize =
        size.width >= size.height ? size.height : size.width;
    int _dimension = _minSize ~/ _noOfModules;
    if (module != null) {
      _dimension = module;
    }
    final double _actualSize = (_noOfModules * _dimension).toDouble();
    final int _xPosition = (offset.dx + (size.width - _actualSize) / 2).toInt();
    int _yPosition = (offset.dy + (size.height - _actualSize) / 2).toInt();

    for (int i = 0; i < w; i++) {
      x = _xPosition;
      for (int j = 0; j < h; j++) {
        if (_moduleValues[i][j]._isBlack) {
          _paint.color = foregroundColor;
        } else {
          _paint.color = Colors.transparent;
        }

        if (_dataAllocationValues != null &&
            _dataAllocationValues[j][i]._isFilled) {
          if (_dataAllocationValues[j][i]._isBlack) {
            _paint.color = foregroundColor;
          }
        }

        final Rect _rect = Rect.fromLTRB(x.toDouble(), _yPosition.toDouble(),
            (x + _dimension).toDouble(), (_yPosition + _dimension).toDouble());
        canvas.drawRect(_rect, _paint);

        x = (x + _dimension).toInt();
      }

      _yPosition = (_yPosition + _dimension).toInt();
    }

    if (showValue) {
      final Offset _offset = Offset(offset.dx, _yPosition.toDouble());
      _drawText(
          canvas, _offset, size, value, textStyle, textSpacing, textAlign);
    }
  }

  /// Method to render the input value of the barcode
  void _drawText(Canvas canvas, Offset offset, Size size, String value,
      TextStyle textStyle, double textSpacing, TextAlign textAlign,
      [Offset _actualOffset, Size _actualSize]) {
    final TextSpan _span = TextSpan(text: value, style: textStyle);

    final TextPainter _textPainter = TextPainter(
        maxLines: 1,
        ellipsis: '.....',
        text: _span,
        textDirection: TextDirection.ltr,
        textAlign: textAlign);
    _textPainter.layout(minWidth: 0, maxWidth: size.width);
    double _x;
    double _y;
    switch (textAlign) {
      case TextAlign.justify:
      case TextAlign.center:
        {
          _x = offset.dx + (size.width / 2 - _textPainter.width / 2);
          _y = offset.dy + textSpacing;
        }
        break;
      case TextAlign.left:
      case TextAlign.start:
        {
          _x = offset.dx;
          _y = offset.dy + textSpacing;
        }
        break;
      case TextAlign.right:
      case TextAlign.end:
        {
          _x = offset.dx + (size.width - _textPainter.width);
          _y = offset.dy + textSpacing;
        }
        break;
    }
    _textPainter.paint(canvas, Offset(_x, _y));
  }
}
